import abc
import datetime as dt
from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from enum import Enum
from typing import Any, Generic, Literal, TypeVar, cast

from pydantic import BaseModel, Field, field_serializer, field_validator

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.registry import Settings
from bayesline.api._src.settings_types import SETTINGS_TYPES

_SETTINGS_TYPES_INDEX = {t.__name__: t for t in SETTINGS_TYPES}

T = TypeVar("T", bound=Settings)


class CouldNotSaveSettingsError(Exception):
    """Raised when a settings object could not be saved to the registry."""


class CouldNotDeleteSettingsError(Exception):
    """Raised when a settings object could not be deleted from the registry."""


class CouldNotPermissionSettingsError(Exception):
    """Raised when a settings object could not be permissioned."""


class SettingsRegistryError(Exception):
    """Raised when an unexpected error occurs in the settings registry."""


@dataclass(frozen=True)
class SettingNotFound:
    """A settings result that could not be found."""

    name: str | None
    id: int | None
    kind: Literal["not_found"] = "not_found"


@dataclass(frozen=True)
class SettingInvalid:
    """
    A settings result that exists but is invalid.

    Possible reasons for invalidity:
    - the underlying json is invalid
    - the setting has references to settings that do not exist
    - the setting has references to settings that are invalid
    - the setting does not validate against the settings menu
    """

    name: str
    id: int
    owner: str
    raw_json: Mapping[str, Any]
    issues: Sequence[str]
    missing_refs: Sequence[str | int] = ()
    kind: Literal["invalid"] = "invalid"


@dataclass(frozen=True)
class SettingNotPermissioned:
    """A settings result that exists but the current user does not have read permissions."""

    name: str
    id: int
    owner: str
    kind: Literal["forbidden"] = "forbidden"


@dataclass(frozen=True)
class SettingValid(Generic[T]):
    """A settings result that exists and is valid."""

    name: str
    id: int
    owner: str
    value: T
    created_on: dt.datetime
    last_updated_on: dt.datetime
    kind: Literal["valid"] = "valid"


SettingResult = (
    SettingValid[T] | SettingInvalid | SettingNotFound | SettingNotPermissioned
)


class ItemStatus(str, Enum):
    """The status of a single item operation, e.g. delete on a single settings object."""

    # order is important for error message construction
    not_found = "not_found"
    invalid_type = "invalid_type"
    forbidden = "forbidden"
    user_not_found = "user_not_found"
    group_not_found = "group_not_found"
    invalid = "invalid"
    conflict = "conflict"
    success = "success"


class ItemResult(BaseModel, extra="forbid", frozen=True):
    """The result of a single item operation, e.g. delete on a single settings object."""

    name_or_id: str | int = Field(
        description=(
            "The name or identifier of the item as it was provided to the operation."
        ),
    )
    status: ItemStatus = Field(description=("The status of the item operation."))
    message: str = Field(
        description="The human readable message of the item's operation result.",
        default="",
    )
    name: str | None = Field(
        description=(
            "The name of the item if available. Will always be available if the "
            "operation was successful."
        )
    )
    id: int | None = Field(
        description=(
            "The identifier of the item if available. Will always be available if the "
            "operation was successful."
        )
    )
    short_message: str = Field(
        description=(
            "A short message of the item's operation result which can be used for "
            "summary messages. E.g. if a user could not be found, the short message "
            "should be the input of the user that could not be found."
        ),
        default="",
    )

    @property
    def success(self) -> bool:
        """Whether the item operation was successful."""
        return self.status == ItemStatus.success


class SaveItemResult(ItemResult):
    """The result of a single save operation."""

    # nothing yet
    pass


class DeleteItemResult(ItemResult):
    """The result of a single delete operation."""

    # nothing yet
    pass


class PermissionItemResult(ItemResult):
    """The result of a single permission operation."""

    # nothing yet
    pass


E = TypeVar("E", bound=ItemResult)


class SettingsRegistryResult(BaseModel, Generic[E], extra="forbid", frozen=True):
    """The result of a batch operation, e.g. a batch delete operation."""

    operation: str = Field(
        description=(
            "The operation that was performed, e.g. 'delete', 'permission', etc."
        ),
    )
    items: list[E] = Field(
        description=(
            "Results of the operations for each settings object in the order they "
            "were provided."
        ),
    )

    @property
    def success(self) -> bool:
        """Whether the entire batch operation was successful."""
        return all(item.success for item in self.items)

    @property
    def name_to_id(self) -> Mapping[str, int]:
        """Mapping of names to identifiers."""
        return {
            item.name: item.id
            for item in self.items
            if item.name is not None and item.id is not None
        }

    @property
    def id_to_name(self) -> Mapping[int, str]:
        """Mapping of identifiers to names."""
        return {
            item.id: item.name
            for item in self.items
            if item.name is not None and item.id is not None
        }

    @property
    def error_message(self) -> str:
        """
        Produces a full error message of the batch operation.

        Will return an empty string if the operation was successful.

        Returns
        -------
        str
            The full error message of the batch operation if the operation was not
            successful or an empty string if the operation was successful.
        """
        if not self.success:
            # collect grouped errors
            grouped_errors: defaultdict[ItemStatus, list[int]] = defaultdict(list)
            for i, item in enumerate(self.items):
                if not item.success:
                    grouped_errors[item.status].append(i)
            msg = f"Could not {self.operation} the following settings: "
            for status, indices in grouped_errors.items():
                messages = [
                    (
                        f"{self.items[i].name_or_id} ({self.items[i].short_message})"
                        if self.items[i].short_message
                        else str(self.items[i].name_or_id)
                    )
                    for i in indices
                ]
                msg += f"\n{status.value}: {', '.join(messages)}"
            return msg
        return ""

    def __getitem__(self, index: int) -> E:  # noqa: D105
        return self.items[index]

    def __len__(self) -> int:  # noqa: D105
        return len(self.items)


class SaveResult(SettingsRegistryResult[SaveItemResult]):
    """The result of a batch save operation."""

    # nothing yet
    pass


class DeleteResult(SettingsRegistryResult[DeleteItemResult]):
    """The result of a batch delete operation."""

    # nothing yet
    pass


class PermissionResult(SettingsRegistryResult[PermissionItemResult]):
    """The result of a batch permission operation."""

    # nothing yet
    pass


class SettingsTransferObject(BaseModel):
    """A settings transfer object."""

    kind: str
    name: str | None
    identifier: int | None
    value: dict[str, Any] = {}
    settings_type: str | None = None
    owner: str | None = None
    issues: Sequence[str] = []
    missing_refs: Sequence[str | int] = []
    created_on: dt.datetime | None = None
    last_updated_on: dt.datetime | None = None

    def to_settings(self, settings_types: list[type[Settings]]) -> SettingResult:
        """
        Convert the settings transfer object to a settings object.

        Parameters
        ----------
        settings_types : list[type[Settings]]
            The available types to use to convert the settings transfer object to.

        Raises
        ------
        ValueError
            If the settings type is not found.
        AssertionError
            If the settings transfer object is not of a known kind.

        Returns
        -------
        SettingResult
            The settings object.

        """
        if self.kind == "not_found":
            return SettingNotFound(name=self.name, id=self.identifier)
        elif self.kind == "invalid":
            return SettingInvalid(
                name=cast(str, self.name),
                id=cast(int, self.identifier),
                owner=cast(str, self.owner),
                raw_json=self.value,
                issues=self.issues,
                missing_refs=self.missing_refs,
            )
        elif self.kind == "forbidden":
            return SettingNotPermissioned(
                name=cast(str, self.name),
                id=cast(int, self.identifier),
                owner=cast(str, self.owner),
            )
        elif self.kind == "valid":
            for t in settings_types:
                if t.__name__ == self.settings_type:
                    break
            else:
                raise ValueError(f"Settings type {self.settings_type} not found")

            return SettingValid(
                name=cast(str, self.name),
                id=cast(int, self.identifier),
                owner=cast(str, self.owner),
                value=t.model_validate(self.value),
                created_on=cast(dt.datetime, self.created_on),
                last_updated_on=cast(dt.datetime, self.last_updated_on),
            )
        else:
            raise AssertionError(f"unknown kind {self.kind}")

    @classmethod
    def from_settings(cls, s: SettingResult) -> "SettingsTransferObject":
        """
        Convert the settings object to a settings transfer object.

        Parameters
        ----------
        s : SettingResult
            The settings object to convert.

        Raises
        ------
        AssertionError
            If the settings object is not of a known kind.

        Returns
        -------
        SettingsTransferObject
            The settings transfer object.
        """
        if s.kind == "not_found":
            return SettingsTransferObject(
                kind=s.kind,
                name=s.name,
                identifier=s.id,
            )
        elif s.kind == "invalid":
            return SettingsTransferObject(
                kind=s.kind,
                name=s.name,
                identifier=s.id,
                owner=s.owner,
                value=s.raw_json,
                issues=s.issues,
                missing_refs=s.missing_refs,
            )
        elif s.kind == "forbidden":
            return SettingsTransferObject(
                kind=s.kind,
                name=s.name,
                identifier=s.id,
                owner=s.owner,
            )
        elif s.kind == "valid":
            return SettingsTransferObject(
                kind=s.kind,
                name=s.name,
                identifier=s.id,
                settings_type=s.value.__class__.__name__,
                owner=s.owner,
                value=s.value.model_dump(),
                created_on=s.created_on,
                last_updated_on=s.last_updated_on,
            )
        else:
            raise AssertionError(f"unknown kind {s.kind}")


class SettingsIdentifiers(BaseModel):
    """Holds mappings of globally unique settings identifiers to other properties."""

    id_to_name: Mapping[int, str] = Field(
        description="Mapping of globally unique identifiers to their names.",
    )
    name_to_id: Mapping[str, int] = Field(
        description="Mapping of names to their globally unique identifiers.",
    )
    id_to_type: Mapping[int, type[Settings]] = Field(
        description="Mapping of globally unique identifiers to the settings type.",
    )
    id_to_user_email: Mapping[int, str] = Field(
        description=(
            "Mapping of globally unique identifiers to the user emails who own them."
        ),
    )

    def __getitem__(self, settings_type: type[Settings]) -> "SettingsIdentifiers":
        """
        Get the identifiers for a specific settings type.

        Parameters
        ----------
        settings_type : type[Settings]
            The settings type to get the identifiers for.

        Raises
        ------
        KeyError
            If the settings type is not covered in this registry.

        Returns
        -------
        SettingsIdentifiers
            The identifiers for the given settings type.
        """
        if not any(t.__name__ == settings_type.__name__ for t in SETTINGS_TYPES):
            raise KeyError(f"Settings type {settings_type.__name__} not found")
        ids_for_type = {k for k, v in self.id_to_type.items() if v == settings_type}
        return SettingsIdentifiers(
            id_to_name={k: v for k, v in self.id_to_name.items() if k in ids_for_type},
            name_to_id={k: v for k, v in self.name_to_id.items() if v in ids_for_type},
            id_to_type={k: v for k, v in self.id_to_type.items() if k in ids_for_type},
            id_to_user_email={
                k: v for k, v in self.id_to_user_email.items() if k in ids_for_type
            },
        )

    @field_validator("id_to_type", mode="before")
    @classmethod
    def parse_type(cls, v: Any) -> dict[int, type[Settings]]:
        if isinstance(v, dict):
            return {
                int(id): (
                    _SETTINGS_TYPES_INDEX[type_] if isinstance(type_, str) else type_
                )
                for id, type_ in v.items()
            }
        raise TypeError("id_to_type must be a dict of int to settings types")

    @field_serializer("id_to_type")
    def serialize_type(self, v: dict[int, type[Settings]]) -> dict[int, str]:
        return {id: type_.__name__ for id, type_ in v.items()}


class TypedSettingsRegistry(abc.ABC, Generic[T]):
    """
    Abstract base class for typed settings registries.

    A typed settings registry covers exactly one settings type, e.g. `UniverseSettings`.
    """

    @property
    @abc.abstractmethod
    def settings_type(self) -> type[T]:
        """
        Get the settings type covered by this registry.

        Returns
        -------
        type[T]
            The settings type covered by this registry.
        """

    @abc.abstractmethod
    def get_identifiers(
        self, which: Literal["all", "mine"] = "all"
    ) -> SettingsIdentifiers:
        """
        Get all identifiers for the settings type covered by this registry.

        Parameters
        ----------
        which : Literal["all", "mine"], default="all"
            'mine' will only return those identfiers owned by the current user
            'all' will also return those identifiers that the current user is
            permissioned to READ.

        Returns
        -------
        SettingsIdentifiers
            All identifiers for the settings type covered by this registry.
        """

    @abc.abstractmethod
    def save(
        self,
        name_or_id: str | int,
        settings: T,
        *,
        overwrite: bool = False,
        raise_on_errors: bool = True,
    ) -> SaveItemResult:
        """
        Save (or update) a settings object to the registry.

        Parameters
        ----------
        name_or_id : str | int
            The name of the settings object to save or the global int id of the settings
            object to update.
            If int id is used, the settings objects must exist in the registry, which
            implies that overwrite must be True.
            The name is a user defined name for the settings object that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be written/updated (if permissioned) by using
            either the global int identifier or prefixing the name with the user name,
            e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
        settings : T
            The settings object to save.
        overwrite : bool, default=False
            If True, overwrite the settings object if it already exists.
            If False, mark an error if the settings object already exists.
        raise_on_errors : bool, default=True
            If True, raise an exception if the settings object could not be saved.
            If False, return a SaveItemResult object with the errors.

        Raises
        ------
        CouldNotSaveSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while saving the settings object.

        Returns
        -------
        SaveItemResult
            Result object for the save operation.
        """

    @abc.abstractmethod
    def save_many(
        self,
        settings: Mapping[str | int, T],
        *,
        overwrite: bool = False,
        raise_on_errors: bool = True,
    ) -> SaveResult:
        """
        Save multiple settings objects to the registry.

        This method is best effort. Partial execution is possible if some settings
        objects can be saved and others cannot.

        Parameters
        ----------
        settings : Mapping[str | int, T]
            The settings objects to save, keys being the names or global int ids of the
            settings objects.
            If int ids are used, the settings objects must exist in the registry, which
            implies that overwrite must be True.
            The name is a user defined name for the settings object that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be written/updated (if permissioned) by using
            either the global int identifier or prefixing the name with the user name,
            e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
        overwrite : bool, default=False
            If True, overwrite the settings objects if they already exist.
            If False, mark an error for settings objects that already exist.
        raise_on_errors : bool, default=True
            If True, raise an exception if any of the settings objects could not be saved.
            If False, return a SaveResult object with the errors.

        Raises
        ------
        CouldNotSaveSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while saving the settings objects.

        Returns
        -------
        SaveResult
            Result object for the save operation.
        """

    @abc.abstractmethod
    def read(self, name_or_id: str | int) -> SettingResult[T]:
        """
        Read a settings object from the registry for the given name or identifier.

        The setting must be of the correct type for this registry.

        Parameters
        ----------
        name_or_id : str | int
            The name or identifier of the settings object to read.
            The identifier is a globally unique integer that is assinged when
            a setting gets first saved to the registry.
            The name is a user defined name for the settings object that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be read (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".

        Returns
        -------
        SettingResult[T]
            The settings object.
        """

    @abc.abstractmethod
    def read_many(
        self, names_or_ids: Sequence[str | int]
    ) -> Sequence[SettingResult[T]]:
        """
        Read settings objects from the registry for the given names or identifiers.

        The settings must be of the correct type for this registry.

        Parameters
        ----------
        names_or_ids: Sequence[str | int]
            The names or identifiers of the settings objects to read.
            The identifiers are globally unique integers that is assinged when
            a setting gets first saved to the registry.
            The names are user defined names for the settings objects that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be read (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".

        Returns
        -------
        Sequence[SettingResult[T]]
            The settings objects in exact same order as the input.
        """

    @abc.abstractmethod
    def delete(
        self, name_or_id: str | int, *, raise_on_errors: bool = True
    ) -> DeleteItemResult:
        """
        Delete a settings object from the registry for the given name or identifier.

        The setting must be of the correct type for this registry.

        For a setting to be able to be deleted, it must exist, not be referenced by
        any other settings and the user must have delete permissions for the setting.

        Parameters
        ----------
        name_or_id : str | int
            The name or identifier of the settings object to delete.
            The identifiers are globally unique integers that is assinged when
            a setting gets first saved to the registry.
            The names are user defined names for the settings objects that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be deleted (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
        raise_on_errors : bool, default=True
            If True, raise an exception if the settings object could not be deleted.
            If False, return a DeleteItemResult object with the errors.

        Raises
        ------
        CouldNotDeleteSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while deleting the settings object.

        Returns
        -------
        DeleteItemResult
            Result object for the delete operation.
        """

    @abc.abstractmethod
    def delete_many(
        self, names_or_ids: Sequence[str | int], *, raise_on_errors: bool = True
    ) -> DeleteResult:
        """
        Delete settings objects from the registry for the given names or identifiers.

        This method is best effort. Partial deletions are possible if some settings
        objects can be deleted and others cannot.

        The settings must be of the correct type for this registry.

        For a setting to be able to be deleted, it must exist, not be referenced by
        any other settings and the user must have delete permissions for the setting.

        Parameters
        ----------
        names_or_ids : Sequence[str | int]
            The names or identifiers of the settings objects to delete.
            The identifier is a globally unique integer that is assinged when
            a setting gets first saved to the registry.
            The name is a user defined name for the settings object that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be deleted (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
        raise_on_errors : bool, default=True
            If True, raise an exception if the settings object could not be deleted
            (any settings that could be deleted will be deleted).
            If False, return a DeleteResult object with the errors.

        Raises
        ------
        CouldNotDeleteSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while deleting the settings objects.

        Returns
        -------
        DeleteResult
            Result object for the delete operation.
        """

    @abc.abstractmethod
    def grant(
        self,
        *,
        permission_to: Literal["read", "write", "delete", "admin"],
        for_setting: str | int,
        to_groups: list[str] | None = None,
        to_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionItemResult:
        """
        Grant a permission to a group or user for given settings.

        The settings must be of the correct type for this registry.

        Parameters
        ----------
        permission_to: Literal["read", "write", "delete", "admin"]
            The type of permission to grant.
            An option includes permission to the preceding options,
            e.g. 'delete' includes 'read', 'write', and 'delete'.
        for_setting: str | int
            The name or identifier of the settings object to grant permissions to.
            The identifier is globally unique integers that is assinged when
            a setting gets first saved to the registry.
            The name is user defined for the settings object that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be used (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
            To grant permissions of settings that belong to another user the grantor
            needs to have either 'admin' permissions for the
            setting in question.
        to_groups: list[str] | None, default=None
            The groups to grant the permission to.
        to_users: list[str] | None, default=None
            The users to grant the permission to (email address).
        including_references: bool, default=False
            If True, grant the permission to the settings objects and all settings
            objects that are referenced by the settings objects.
            If False, grant the permission to the settings objects only.
        raise_on_errors : bool, default=True
            If True, raise an exception if the permissions could not be granted.
            If False, return a PermissionItemResult object with the errors.

        Raises
        ------
        CouldNotPermissionSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while permissioning the settings objects.

        Returns
        -------
        PermissionItemResult
            Result object for the permissioning operation.
        """

    @abc.abstractmethod
    def grant_many(
        self,
        *,
        permission_to: Literal["read", "write", "delete", "admin"],
        for_settings: list[str | int],
        to_groups: list[str] | None = None,
        to_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionResult:
        """
        Grant a permission to a group or user for given settings.

        This method is best effort. Partial execution is possible if some settings
        objects can be granted permissions and others cannot.

        The settings must be of the correct type for this registry.

        Parameters
        ----------
        permission_to: Literal["read", "write", "delete", "admin"]
            The type of permission to grant.
            An option includes permission to the preceding options,
            e.g. 'delete' includes 'read', 'write', and 'delete'.
        for_settings: list[str | int]
            The names or identifiers of the settings objects to read.
            The identifiers are globally unique integers that is assinged when
            a setting gets first saved to the registry.
            The names are user defined names for the settings objects that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be used (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
            To grant permissions of settings that belong to another user the grantor
            needs to have 'admin' permissions for the
            setting in question.
        to_groups: list[str] | None, default=None
            The groups to grant the permission to.
        to_users: list[str] | None, default=None
            The users to grant the permission to (email address).
        including_references: bool, default=False
            If True, grant the permission to the settings objects and all settings
            objects that are referenced by the settings objects.
            If False, grant the permission to the settings objects only.
        raise_on_errors : bool, default=True
            If True, raise an exception if the permissions could not be granted.
            If False, return a PermissionResult object with the errors.

        Raises
        ------
        CouldNotPermissionSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while permissioning the settings objects.

        Returns
        -------
        PermissionResult
            Result object for the permissioning operation.
        """

    @abc.abstractmethod
    def revoke(
        self,
        *,
        for_setting: str | int,
        from_groups: list[str] | None = None,
        from_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionItemResult:
        """
        Revoke a permission from a group or user for given settings.

        The settings must be of the correct type for this registry.

        Parameters
        ----------
        for_setting: str | int
            The names or identifiers of the settings objects to read.
            The identifiers are globally unique integers that is assinged when
            a setting gets first saved to the registry.
            The names are user defined names for the settings objects that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be used (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
            To revoke permissions of settings that belong to another user the
            revoking user needs to have 'admin' permissions
            for the setting in question.
        from_groups: list[str] | None, default=None
            The groups to revoke the permission from.
        from_users: list[str] | None, default=None
            The users to revoke the permission from (email address).
        including_references: bool, default=False
            If True, revoke the permission from the settings objects and all settings
            objects that are referenced by the settings objects.
            If False, revoke the permission from the settings objects only.
        raise_on_errors : bool, default=True
            If True, raise an exception if the permissions could not be revoked.
            If False, return a PermissionItemResult object with the errors.

        Raises
        ------
        CouldNotPermissionSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while permissioning the settings objects.

        Returns
        -------
        PermissionItemResult
            Result object for the permissioning operation.
        """

    @abc.abstractmethod
    def revoke_many(
        self,
        *,
        for_settings: list[str | int],
        from_groups: list[str] | None = None,
        from_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionResult:
        """
        Revoke a permission from a group or user for given settings.

        This method is best effort. Partial execution is possible if some settings
        objects can be revoked permissions from and others cannot.

        The settings must be of the correct type for this registry.

        Parameters
        ----------
        for_settings: list[str | int]
            The names or identifiers of the settings objects to read.
            The identifiers are globally unique integers that is assinged when
            a setting gets first saved to the registry.
            The names are user defined names for the settings objects that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be used (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
            To revoke permissions of settings that belong to another user the
            revoking user needs to have 'admin' permissions
            for the setting in question.
        from_groups: list[str] | None, default=None
            The groups to revoke the permission from.
        from_users: list[str] | None, default=None
            The users to revoke the permission from (email address).
        including_references: bool, default=False
            If True, revoke the permission from the settings objects and all settings
            objects that are referenced by the settings objects.
            If False, revoke the permission from the settings objects only.
        raise_on_errors : bool, default=True
            If True, raise an exception if the permissions could not be revoked.
            If False, return a PermissionResult object with the errors.

        Raises
        ------
        CouldNotPermissionSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while permissioning the settings objects.

        Returns
        -------
        PermissionResult
            Result object for the permissioning operation.
        """

    @abc.abstractmethod
    def list_user_names(self) -> Mapping[str, str]:
        """
        List all users that settings can be granted permissions to.

        Returns
        -------
        Mapping[str, str]
            Mapping of unique user id (e.g. email) to their names.
        """


class SettingsRegistry(abc.ABC):
    """
    Abstract base class for settings registries.

    This settings registry covers all available settings types, e.g. `UniverseSettings`,
    `PortfolioSettings`, etc.
    """

    def get_settings_types(self) -> Sequence[type[Settings]]:
        """
        Get all settings types covered in this registry.

        Returns
        -------
        Sequence[type[Settings]]
            A sequence of the settings types covered in this registry.
        """
        return SETTINGS_TYPES

    @abc.abstractmethod
    def __getitem__(self, settings_type: type[T]) -> TypedSettingsRegistry[T]:
        """
        Get the typed settings registry for the given settings type.

        Parameters
        ----------
        settings_type : type[T]
            The settings type to get the registry for.

        Raises
        ------
        KeyError
            If the settings type is not covered in this registry.

        Returns
        -------
        TypedSettingsRegistry[T]
            The typed settings registry for the given settings type.
        """

    @abc.abstractmethod
    def get_identifiers(
        self, which: Literal["all", "mine"] = "all"
    ) -> SettingsIdentifiers:
        """
        Get all identifiers for all settings types covered in this registry.

        Parameters
        ----------
        which : Literal["all", "mine"], default="all"
            'mine' will only return those identfiers owned by the current user
            'all' will also return those identifiers that the current user is
            permissioned to READ.

        Returns
        -------
        SettingsIdentifiers
            The identifiers for all settings types covered in this registry.
        """

    @abc.abstractmethod
    def save_many(
        self,
        settings: Mapping[str | int, Settings],
        *,
        overwrite: bool = False,
        raise_on_errors: bool = True,
    ) -> SaveResult:
        """
        Save multiple settings objects to the registry.

        This method is best effort. Partial execution is possible if some settings
        objects can be saved and others cannot.

        Parameters
        ----------
        settings : Mapping[str | int, Settings]
            The settings objects to save, keys being the names or global int ids of the
            settings objects.
            If int ids are used, the settings objects must exist in the registry, which
            implies that overwrite must be True.
            The name is a user defined name for the settings object that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be written/updated (if permissioned) by using
            either the global int identifier or prefixing the name with the user name,
            e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
        overwrite : bool, default=False
            If True, overwrite the settings objects if they already exist.
            If False, mark an error for settings objects that already exist.
        raise_on_errors : bool, default=True
            If True, raise an exception if any of the settings objects could not be saved.
            If False, return a SaveResult object with the errors.

        Raises
        ------
        CouldNotSaveSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while saving the settings objects.

        Returns
        -------
        SaveResult
            Result object for the save operation.
        """

    @abc.abstractmethod
    def read_many(
        self, names_or_ids: Sequence[str | int]
    ) -> Sequence[SettingResult[Settings[Any]]]:
        """
        Read settings objects from the registry for the given names or identifiers.

        Parameters
        ----------
        names_or_ids: Sequence[str | int]
            The names or identifiers of the settings objects to read.
            The identifiers are globally unique integers that is assinged when
            a setting gets first saved to the registry.
            The names are user defined names for the settings objects that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be read (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".

        Returns
        -------
        Sequence[SettingResult[Settings[Any]]]
            The settings objects in exact same order as the input.
        """

    @abc.abstractmethod
    def delete_many(
        self, names_or_ids: Sequence[str | int], *, raise_on_errors: bool = True
    ) -> DeleteResult:
        """
        Delete settings objects from the registry for the given names or identifiers.

        This method is best effort. Partial deletions are possible if some settings
        objects can be deleted and others cannot.

        For a setting to be able to be deleted, it must exist, not be referenced by
        any other settings and the user must have delete permissions for the setting.

        Parameters
        ----------
        names_or_ids : Sequence[str | int]
            The names or identifiers of the settings objects to delete.
            The identifier is a globally unique integer that is assinged when
            a setting gets first saved to the registry.
            The name is a user defined name for the settings object that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be deleted (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
        raise_on_errors : bool, default=True
            If True, raise an exception if the settings object could not be deleted
            (any settings that could be deleted will be deleted).
            If False, return a DeleteResult object with the errors.

        Raises
        ------
        CouldNotDeleteSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while deleting the settings objects.

        Returns
        -------
        DeleteResult
            Result object for the delete operation.
        """

    @abc.abstractmethod
    def grant_many(
        self,
        *,
        permission_to: Literal["read", "write", "delete", "admin"],
        for_settings: list[str | int],
        to_groups: list[str] | None = None,
        to_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionResult:
        """
        Grant a permission to a group or user for given settings.

        This method is best effort. Partial execution is possible if some settings
        objects can be granted permissions and others cannot.

        Parameters
        ----------
        permission_to: Literal["read", "write", "delete", "admin"]
            The type of permission to grant.
            An option includes permission to the preceding options,
            e.g. 'delete' includes 'read', 'write', and 'delete'.
        for_settings: list[str | int]
            The names or identifiers of the settings objects to read.
            The identifiers are globally unique integers that is assinged when
            a setting gets first saved to the registry.
            The names are user defined names for the settings objects that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be used (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
            To grant permissions of settings that belong to another user the grantor
            needs to have 'admin' permissions for the
            setting in question.
        to_groups: list[str] | None, default=None
            The groups to grant the permission to.
        to_users: list[str] | None, default=None
            The users to grant the permission to (email address).
        including_references: bool, default=False
            If True, grant the permission to the settings objects and all settings
            objects that are referenced by the settings objects.
            If False, grant the permission to the settings objects only.
        raise_on_errors: bool, default=True
            If True, raise an exception if the permissions could not be granted.
            If False, return a PermissionResult object with the errors.

        Raises
        ------
        CouldNotPermissionSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while permissioning the settings objects.

        Returns
        -------
        PermissionResult
            Result objects for the permissioning operation.
        """

    @abc.abstractmethod
    def revoke_many(
        self,
        *,
        for_settings: list[str | int],
        from_groups: list[str] | None = None,
        from_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionResult:
        """
        Revoke a permission from a group or user for given settings.

        This method is best effort. Partial execution is possible if some settings
        objects can be revoked permissions from and others cannot.

        Parameters
        ----------
        for_settings: list[str | int]
            The names or identifiers of the settings objects to read.
            The identifiers are globally unique integers that is assinged when
            a setting gets first saved to the registry.
            The names are user defined names for the settings objects that must
            consist of only alphanumeric characters, underscores and hyphens and start
            with a letter.
            Other users' settings can be used (if permissioned) by using either
            the global int identifier or prefixing the name with the user name and
            a forward slash, e.g. "sebastian.janisch@bayesline.com/SebastiansSettings".
            To revoke permissions of settings that belong to another user the
            revoking user needs to have 'admin' permissions
            for the setting in question.
        from_groups: list[str] | None, default=None
            The groups to revoke the permission from.
        from_users: list[str] | None, default=None
            The users to revoke the permission from (email address).
        including_references: bool, default=False
            If True, revoke the permission from the settings objects and all settings
            objects that are referenced by the settings objects.
            If False, revoke the permission from the settings objects only.
        raise_on_errors: bool, default=True
            If True, raise an exception if the permissions could not be revoked.
            If False, return a PermissionResult object with the errors.

        Raises
        ------
        CouldNotPermissionSettingsError
            If an error captured by the item result is encountered and raise_on_errors
            is True
        SettingsRegistryError
            If an unexpected error occurs while permissioning the settings objects.

        Returns
        -------
        PermissionResult
            Result objects for the permissioning operation.
        """

    @abc.abstractmethod
    def list_user_names(self) -> Mapping[str, str]:
        """
        List all users that settings can be granted permissions to.

        Returns
        -------
        Mapping[str, str]
            Mapping of unique user id (e.g. email) to their names.
        """


@docstrings_from_sync
class AsyncTypedSettingsRegistry(abc.ABC, Generic[T]):

    @property
    @abc.abstractmethod
    def settings_type(self) -> type[T]: ...  # noqa: D102

    @abc.abstractmethod
    async def get_identifiers(  # noqa: D102
        self, which: Literal["all", "mine"] = "all"
    ) -> SettingsIdentifiers: ...

    @abc.abstractmethod
    async def save(  # noqa: D102
        self,
        name_or_id: str | int,
        settings: T,
        *,
        overwrite: bool = False,
        raise_on_errors: bool = True,
    ) -> SaveItemResult: ...

    @abc.abstractmethod
    async def save_many(  # noqa: D102
        self,
        settings: Mapping[str | int, T],
        *,
        overwrite: bool = False,
        raise_on_errors: bool = True,
    ) -> SaveResult: ...

    @abc.abstractmethod
    async def read(self, name_or_id: str | int) -> SettingResult[T]: ...  # noqa: D102

    @abc.abstractmethod
    async def read_many(  # noqa: D102
        self, names_or_ids: Sequence[str | int]
    ) -> Sequence[SettingResult[T]]: ...

    @abc.abstractmethod
    async def delete(  # noqa: D102
        self, name_or_id: str | int, *, raise_on_errors: bool = True
    ) -> DeleteItemResult: ...

    @abc.abstractmethod
    async def delete_many(  # noqa: D102
        self, names_or_ids: Sequence[str | int], *, raise_on_errors: bool = True
    ) -> DeleteResult: ...

    @abc.abstractmethod
    async def grant(  # noqa: D102
        self,
        *,
        permission_to: Literal["read", "write", "delete", "admin"],
        for_setting: str | int,
        to_groups: list[str] | None = None,
        to_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionItemResult: ...

    @abc.abstractmethod
    async def grant_many(  # noqa: D102
        self,
        *,
        permission_to: Literal["read", "write", "delete", "admin"],
        for_settings: list[str | int],
        to_groups: list[str] | None = None,
        to_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionResult: ...

    @abc.abstractmethod
    async def revoke(  # noqa: D102
        self,
        *,
        for_setting: str | int,
        from_groups: list[str] | None = None,
        from_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionItemResult: ...

    @abc.abstractmethod
    async def revoke_many(  # noqa: D102
        self,
        *,
        for_settings: list[str | int],
        from_groups: list[str] | None = None,
        from_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionResult: ...

    @abc.abstractmethod
    async def list_user_names(self) -> Mapping[str, str]: ...  # noqa: D102


@docstrings_from_sync
class AsyncSettingsRegistry(abc.ABC):

    async def get_settings_types(self) -> Sequence[type[Settings]]:
        """
        Get all settings types covered in this registry.

        Returns
        -------
        Sequence[type[Settings]]
            A sequence of the settings types covered in this registry.
        """
        return SETTINGS_TYPES

    @abc.abstractmethod
    def __getitem__(  # noqa: D105
        self, settings_type: type[T]
    ) -> AsyncTypedSettingsRegistry[T]: ...

    @abc.abstractmethod
    async def get_identifiers(  # noqa: D102
        self, which: Literal["all", "mine"] = "all"
    ) -> SettingsIdentifiers: ...

    @abc.abstractmethod
    async def save_many(  # noqa: D102
        self,
        settings: Mapping[str | int, Settings],
        *,
        overwrite: bool = False,
        raise_on_errors: bool = True,
    ) -> SaveResult: ...

    @abc.abstractmethod
    async def read_many(  # noqa: D102
        self, names_or_ids: Sequence[str | int]
    ) -> Sequence[SettingResult[Settings[Any]]]: ...

    @abc.abstractmethod
    async def delete_many(  # noqa: D102
        self, names_or_ids: Sequence[str | int], *, raise_on_errors: bool = True
    ) -> DeleteResult: ...

    @abc.abstractmethod
    async def grant_many(  # noqa: D102
        self,
        *,
        permission_to: Literal["read", "write", "delete", "admin"],
        for_settings: list[str | int],
        to_groups: list[str] | None = None,
        to_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionResult: ...

    @abc.abstractmethod
    async def revoke_many(  # noqa: D102
        self,
        *,
        for_settings: list[str | int],
        from_groups: list[str] | None = None,
        from_users: list[str] | None = None,
        including_references: bool = False,
        raise_on_errors: bool = True,
    ) -> PermissionResult: ...

    @abc.abstractmethod
    async def list_user_names(self) -> Mapping[str, str]: ...  # noqa: D102



import abc
from typing import ClassVar, Self, Sequence

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    ReportAccessorApi,
)
from bayesline.api._src.equity.report.scenarios import AnyScenarioSettings


class TypedReportAccessorApi(abc.ABC):
    """
    A base interface for typed report accessor APIs.

    It is meant to be extended by concrete report implementations, e.g.
    a `FactorCovarianceReportAccessorApi` which adds concrete functions that
    operate on the underyling low level report accessor and slice report data
    to specific representations, e.g. `def fcov` which returns the correct
    dataframe representation of the factor covariance matrix instead the raw
    report dataframe.
    """

    scenario_types: ClassVar[list[type[AnyScenarioSettings]]] = []

    @abc.abstractmethod
    def __init__(self, accessor: ReportAccessorApi): ...  # noqa: DOC101 DOC103

    @property
    @abc.abstractmethod
    def accessor(self) -> ReportAccessorApi:
        """
        The underlying report accessor API.

        Returns
        -------
        ReportAccessorApi
            The underlying report accessor API.
        """

    def with_scenario(
        self, scenario_settings: AnyScenarioSettings | Sequence[AnyScenarioSettings]
    ) -> Self:
        """Return a new accessor with the given scenario settings applied.

        Parameters
        ----------
        scenario_settings : AnyScenarioSettings | Sequence[AnyScenarioSettings]
            The scenario settings to apply.

        Returns
        -------
        Self
            A new accessor with the scenario settings applied.

        Raises
        ------
        NotImplementedError
            If this accessor does not support scenarios.
        """
        new_inner = self.accessor.with_scenario_settings(scenario_settings)
        return type(self)(new_inner)


@docstrings_from_sync
class AsyncTypedReportAccessorApi(abc.ABC):

    scenario_types: ClassVar[list[type[AnyScenarioSettings]]] = []

    @abc.abstractmethod
    def __init__(self, accessor: AsyncReportAccessorApi): ...

    @property
    @abc.abstractmethod
    def accessor(self) -> AsyncReportAccessorApi: ...  # noqa: D102

    async def with_scenario(  # noqa: D102
        self, scenario_settings: AnyScenarioSettings | Sequence[AnyScenarioSettings]
    ) -> Self:
        new_inner = await self.accessor.with_scenario_settings(scenario_settings)
        return type(self)(new_inner)



from __future__ import annotations

import json
import os
from collections.abc import Mapping
from typing import Annotated, Any, Literal, overload

import polars as pl
from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)

from bayesline.api._src.equity import settings as settings_tools
from bayesline.api._src.equity.calendar_settings import CalendarSettingsMenu
from bayesline.api._src.equity.universe_settings import (
    CategoricalFilterSettings,
    UniverseSettings,
    UniverseSettingsMenu,
)
from bayesline.api._src.registry import Settings, SettingsMenu

Hierarchy = settings_tools.Hierarchy


def _ensure_list(value: Any) -> Any:
    if isinstance(value, list | tuple):
        return value
    return [value]


class HierarchyDescription(BaseModel, frozen=True, extra="forbid"):
    """Base class for hierarchy descriptions."""

    hierarchy_type: Literal["level", "groups"]
    name: str = Field(
        min_length=1,
        description="""
        The name of the hierarchy to use, e.g. 'trbc' or 'continent'.
        If it is not given then the default hierarchy will be used.
        """,
        examples=["trbc", "continent"],
    )

    def validate_hierarchy_description(self, menu: ExposureSettingsMenu | None) -> None:
        """Validate the hierarchy description.

        Parameters
        ----------
        menu : ExposureSettingsMenu | None
            The menu to validate against. If None, the hierarchy description is validated
            without having the menu context. This means that less validation is done.
        """
        self.model_validate(self.model_dump(), context=menu)


class HierarchyLevel(HierarchyDescription):
    """The hierarchy decscription for a level in the hierarchy.

    E.g. for industries specifying level `1` would
    create top level sector factors.
    """

    hierarchy_type: Literal["level"] = "level"
    level: int = Field(
        description="""The level of the hierarchy to use, e.g. 1
        to use all level 1 names (i.e. sectors for industries or
        continents for regions) or 2 to use all level 2
        names (i.e. sub-sectors for industries and
        countries for regions).
        """,
        default=1,
        examples=[1, 2],
        ge=1,
    )

    @field_validator("level", mode="after")
    @classmethod
    def check_valid_level(cls, level: int, info: ValidationInfo) -> int:
        if isinstance(menu := info.context, ExposureSettingsMenu):
            if (name := info.data["name"]) not in menu.hierarchies:
                raise ValueError(f"Hierarchy {name} does not exist")
            max_level = settings_tools.get_depth(menu.hierarchies[name])
            if level > max_level:
                raise ValueError(f"Illegal level {level}, maximum level is {max_level}")
        return level


class HierarchyGroups(HierarchyDescription):
    """The hierarchy decscription for a custom nested grouping of the hierarchy.

    The top level groupings will turn into factors, whereas any nested
    groupings will be retained for other uses (e.g. risk decomposition).
    """

    hierarchy_type: Literal["groups"] = "groups"
    groupings: Mapping[str, Hierarchy] = Field(  # at least two levels deep
        description="""
        A nested structure of groupings where the keys are the group names
        and the leaf level is a list of hierarchy codes or labels (at any level)
        to include for this group.
        """,
    )

    @field_validator("groupings", mode="after")
    @classmethod
    def check_valid_groupings(
        cls, groupings: Mapping[str, Hierarchy], info: ValidationInfo
    ) -> Mapping[str, Hierarchy]:
        settings_tools.check_unique_hierarchy(groupings)
        settings_tools.assert_no_empty_branches(groupings)
        leaves = settings_tools.flatten(groupings, only_leaves=True)
        if len(leaves) != len(set(leaves)):
            raise ValueError(
                "Groupings must be unique in the sense that no code or label occurs "
                "under any group name."
            )

        if isinstance(menu := info.context, ExposureSettingsMenu):
            # validation dependent on the hierarchy type
            leaves_without_prefix = [  # strip the ~ prefix from the leaves
                g[1:] if g.startswith("~") else g
                for g in settings_tools.flatten(groupings, only_leaves=True)
            ]
            if non_leaves_with_prefix := [
                g
                for g in (
                    set(settings_tools.flatten(groupings))
                    - set(settings_tools.flatten(groupings, only_leaves=True))
                )
                if g.startswith("~")
            ]:
                raise ValueError(
                    "Only leaves can have the exlude prefix ~. The following "
                    f"non-leaves have a ~ prefix: {', '.join(non_leaves_with_prefix)}."
                )
            if (name := info.data["name"]) not in menu.hierarchies:
                raise ValueError(f"Hierarchy {name} does not exist")
            effective_codes = settings_tools.labels_to_codes(  # prefix not resolved
                leaves_without_prefix, menu.hierarchies_labels[name]
            )
            available_codes = settings_tools.flatten(
                menu.hierarchies[name], only_leaves=False
            )
            if missing := set(effective_codes) - set(available_codes):
                raise ValueError(
                    "The following codes are in the hierarchy but not in the hierarchy "
                    f"groups: {', '.join(missing)}"
                )
        return groupings


def _hierarchy_name_to_hierarchy_level(
    v: HierarchyLevel | HierarchyGroups | str,
) -> HierarchyLevel | HierarchyGroups:
    # accept strings and turn them into a level-1 HierarchyLevel
    if isinstance(v, str):
        return HierarchyLevel(name=v, level=1)
    return v


HierarchyType = Annotated[
    HierarchyLevel | HierarchyGroups,
    Field(discriminator="hierarchy_type"),
    BeforeValidator(_hierarchy_name_to_hierarchy_level),
]


class BaseExposureGroupSettings(BaseModel, frozen=True, extra="forbid"):
    """Base class for exposure group settings."""

    def validate_exposure_group_settings(
        self, menu: ExposureSettingsMenu | None
    ) -> None:
        """Validate the exposure group settings.

        Parameters
        ----------
        menu : ExposureSettingsMenu | None
            The menu to validate against. If None, the exposure group settings is validated
            without having the menu context. This means that less validation is done.
        """
        self.model_validate(self.model_dump(), context=menu)


class ContinuousExposureGroupSettings(BaseExposureGroupSettings):
    """The settings for a continuous exposure group.

    Continuous exposures are exposures that are measured on a continuous scale,
    e.g. market, size, momentum, etc., and are typically available for most assets.
    """

    exposure_type: Literal["continuous"] = "continuous"
    hierarchy: Annotated[
        HierarchyType,
        Field(
            description=(
                "The hierarchy to use for the continuous exposures. This is either a"
                "HierarchyLevel or HierarchyGroups object. If a string is passed, the "
                "value is converted to a HierarchyLevel of that hierarchy with level=1."
            ),
            examples=[
                "market",
                HierarchyLevel(name="style", level=2),
                HierarchyGroups(
                    name="style",
                    groupings={
                        "momentum": ["mom6"],
                        "size": ["size"],
                    },
                ),
            ],
        ),
    ]
    factor_group: str = Field(
        "",  # temporary value that does not pass validation
        description=(
            "By default, the name of the factor group will be the name of the hierarchy. "
            "But we can override this by specifying an alias here."
        ),
        examples=["style", "some_style_group"],
    )
    include: Literal["All"] | list[str] = Field(
        default="All",
        description=(
            "Valid hierarchy codes or labels for given hierarchy at any level. If "
            "labels are used which may be duplicated, then the code with the highest "
            "level is used. If 'All', all codes are included."
        ),
        examples=[["momentum"], "All", ["Size", "value"]],
    )
    exclude: list[str] = Field(
        default_factory=list,
        description=(
            "Valid hierarchy codes or labels for given hierarchy at any level. If "
            "labels are used which may be duplicated, then the code with the lowest "
            "level is used."
        ),
        examples=[["momentum"], ["Size", "value"]],
    )
    standardize_method: Literal["none", "equal_weighted"] = Field(
        "none",
        description=(
            "The method to use for standardizing the exposures. If 'none', no "
            "standardization is applied. If 'equal_weighted', then the exposures are "
            "standardized with the mean and standard deviation of the estimation "
            "universe."
        ),
        examples=["none", "equal_weighted"],
    )

    @model_validator(mode="before")
    @classmethod
    def set_factor_group(cls, data: Any) -> Any:
        if isinstance(data, dict):
            if "factor_group" not in data or data["factor_group"] is None:
                if isinstance(data["hierarchy"], HierarchyLevel | HierarchyGroups):
                    data["factor_group"] = data["hierarchy"].name
                elif isinstance(data["hierarchy"], str):
                    data["factor_group"] = data["hierarchy"]
                else:
                    raise ValueError(
                        f"Invalid hierarchy type: {type(data['hierarchy'])}"
                    )
        return data

    @field_validator("factor_group")
    @classmethod
    def check_valid_factor_group(cls, v: str) -> str:
        if v == "":
            raise ValueError("factor_group cannot be empty")
        return v

    @model_validator(mode="after")
    def validate_filters(self, info: ValidationInfo) -> ContinuousExposureGroupSettings:
        if isinstance(menu := info.context, ExposureSettingsMenu):
            # validate that the hierarchies exist
            settings_tools.validate_hierarchy_schema(
                menu.continuous_hierarchies, self.hierarchy.name
            )
            # common validation
            settings_tools.validate_hierarchy_filters(
                menu.hierarchies[self.hierarchy.name],
                self.include,
                self.exclude,
                menu.hierarchies_labels[self.hierarchy.name],
            )
            # validate the hierarchy
            self.hierarchy.validate_hierarchy_description(menu)
        return self

    @property
    def hierarchies(self) -> list[HierarchyType]:
        """The list of hierarchies for the exposure group.

        Returns
        -------
        list[HierarchyType]
            The list of hierarchies for the exposure group.
        """
        return [self.hierarchy]

    def effective_exposure_group_factors(
        self,
        menu: ExposureSettingsMenu,
    ) -> list[str]:
        """Get the effective factors.

        Effective factors are the factors that are included and not excluded in the
        exposure group settings.

        Parameters
        ----------
        menu : ExposureSettingsMenu
            The menu to get the total factors from.

        Returns
        -------
        list[str]
            The list of effective factors.
        """
        return _effective_exposure_group_factors(self, menu)

    def normalize_group_settings(
        self,
        menu: ExposureSettingsMenu,
        universe_filter_dict: Mapping[str, CategoricalFilterSettings] | None,
    ) -> ContinuousExposureGroupSettings:
        """Normalize the exposure group settings.

        Normalized group settings have all hierarchies replaced by HierarchyGroups with
        two levels, the factor names and the leaf codes. All includes and excludes are
        resolved.

        This means that if the universe_filter_dict has a filter that trims down to just
        the North American region, and the ExposureSettings include factors from the
        same country hierarchy, then the factors will be trimmed down to just include
        the ones corresponding to the North American region.

        Parameters
        ----------
        menu : ExposureSettingsMenu
            The menu to get the total factors from.
        universe_filter_dict : Mapping[str, CategoricalFilterSettings] | None
            The universe filter dictionary to potentially further cut the factors down.
            The keys are the hierarchy names and the values are the filter settings.
            If None, then no filtering is performed.

        Returns
        -------
        ContinuousExposureGroupSettings
            The normalized exposure group settings.
        """
        return _normalize_group_settings(self, menu, universe_filter_dict)


class CategoricalExposureGroupSettings(BaseExposureGroupSettings):
    """The settings for a categorical exposure group.

    Categorical exposures are exposures to a categorical variable, e.g. industry,
    country, etc. For example, an asset exposure may be to a single industry factor,
    and we say the exposure is to the industry category factor "Materials". The exposure
    itself may be continuous (e.g. not 0.0 or 1.0), and an asset may be exposed to
    multiple industry factors (but typically not all of them). The distinction with
    continuous exposures is primarily in the settings that are available.
    """

    exposure_type: Literal["categorical"] = "categorical"
    hierarchy: Annotated[
        HierarchyType,
        Field(
            description=(
                "The hierarchy to use for the categorical exposures. This is either a"
                "HierarchyLevel or HierarchyGroups object. If a string is passed, the "
                "value is converted to a HierarchyLevel of that hierarchy with level=1."
            ),
            examples=[
                "trbc",
                HierarchyLevel(name="trbc", level=2),
                HierarchyGroups(
                    name="style",
                    groupings={
                        "MyGroup1": ["Energy"],
                        "MyGroup2": ["Materials", "Chemicals"],
                    },
                ),
            ],
        ),
    ]
    factor_group: str = Field(
        "",  # temporary value that does not pass validation
        description=(
            "By default, the name of the factor group will be the name of the hierarchy. "
            "But we can override this by specifying an alias here."
        ),
        examples=["industry", "some_country_group"],
    )
    include: Literal["All"] | list[str] = Field(
        default="All",
        description=(
            "Valid industry codes or labels for given hierarchy at any level. If "
            "labels are used which may be duplicated, then the code with the highest "
            "level is used. If 'All', all codes are included."
        ),
        examples=[["3571"], "All", ["Materials", "1010"], ["Europe", "CAN"]],
    )
    exclude: list[str] = Field(
        default_factory=list,
        description=(
            "Valid industry codes or labels for given hierarchy at any level. If "
            "labels are used which may be duplicated, then the code with the lowest "
            "level is used."
        ),
        examples=[["3571"], ["Materials", "1010"], ["JPN"]],
    )

    @model_validator(mode="before")
    @classmethod
    def set_factor_group(cls, data: Any) -> Any:
        if isinstance(data, dict):
            if "factor_group" not in data or data["factor_group"] is None:
                if isinstance(data["hierarchy"], HierarchyLevel | HierarchyGroups):
                    data["factor_group"] = data["hierarchy"].name
                elif isinstance(data["hierarchy"], str):
                    data["factor_group"] = data["hierarchy"]
                else:
                    raise ValueError(
                        f"Invalid hierarchy type: {type(data['hierarchy'])}"
                    )
        return data

    @field_validator("factor_group")
    @classmethod
    def check_valid_factor_group(cls, v: str) -> str:
        if v == "":
            raise ValueError("factor_group cannot be empty")
        return v

    @model_validator(mode="after")
    def validate_filters(
        self, info: ValidationInfo
    ) -> CategoricalExposureGroupSettings:
        if isinstance(menu := info.context, ExposureSettingsMenu):
            # validate that the hierarchies exist
            settings_tools.validate_hierarchy_schema(
                menu.categorical_hierarchies, self.hierarchy.name
            )
            # common validation
            settings_tools.validate_hierarchy_filters(
                menu.hierarchies[self.hierarchy.name],
                self.include,
                self.exclude,
                menu.hierarchies_labels[self.hierarchy.name],
            )
            # validate the hierarchy
            self.hierarchy.validate_hierarchy_description(menu)
        return self

    @property
    def hierarchies(self) -> list[HierarchyType]:
        """The list of hierarchies for the exposure group.

        Returns
        -------
        list[HierarchyType]
            The list of hierarchies for the exposure group.
        """
        return [self.hierarchy]

    def effective_exposure_group_factors(
        self,
        menu: ExposureSettingsMenu,
    ) -> list[str]:
        """Get the effective factors.

        Effective factors are the factors that are included and not excluded in the
        exposure group settings.

        Parameters
        ----------
        menu : ExposureSettingsMenu
            The menu to get the total factors from.

        Returns
        -------
        list[str]
            The list of effective factors.
        """
        return _effective_exposure_group_factors(self, menu)

    def normalize_group_settings(  # same as ContinuousExposureGroupSettings
        self,
        menu: ExposureSettingsMenu,
        universe_filter_dict: Mapping[str, CategoricalFilterSettings] | None,
    ) -> CategoricalExposureGroupSettings:
        """Normalize the exposure group settings.

        Normalized group settings have all hierarchies replaced by HierarchyGroups with
        two levels, the factor names and the leaf codes. All includes and excludes are
        resolved.

        This means that if the universe_filter_dict has a filter that trims down to just
        the North American region, and the ExposureSettings include factors from the
        same country hierarchy, then the factors will be trimmed down to just include
        the ones corresponding to the North American region.

        Parameters
        ----------
        menu : ExposureSettingsMenu
            The menu to get the total factors from.
        universe_filter_dict : Mapping[str, CategoricalFilterSettings] | None
            The universe filter dictionary to potentially further cut the factors down.
            The keys are the hierarchy names and the values are the filter settings.
            If None, then no filtering is performed.

        Returns
        -------
        CategoricalExposureGroupSettings
            The normalized exposure group settings.
        """
        return _normalize_group_settings(self, menu, universe_filter_dict)


class TSBetaExposureGroupSettings(BaseExposureGroupSettings):
    """Exposure group settings for time-series beta exposures from uploaded time-series.

    Computes rolling betas (OLS slopes) between asset returns and uploaded factor
    time-series at report time, without requiring a risk dataset rebuild.
    """

    exposure_type: Literal["tsbeta"] = "tsbeta"
    tsfactors_source: str = Field(
        description="The name of the uploaded time-series factors dataset.",
    )
    factor_group: str = Field(
        description="The output factor group name for these exposures.",
    )
    include: Literal["All"] | list[str] = Field(
        default="All",
        description="Factor names to include from the uploaded dataset. 'All' includes all.",
    )
    exclude: list[str] = Field(
        default_factory=list,
        description="Factor names to exclude from the uploaded dataset.",
    )
    currency: str = Field(
        default="USD",
        description="Currency for computing asset returns.",
    )
    window: int = Field(
        default=252,
        description="Rolling window size for time-series beta computation.",
        ge=2,
    )
    overlap: int = Field(
        default=5,
        description=(
            "Overlap (convolution) window for overlapped returns before beta computation."
        ),
        ge=1,
    )
    gaussianize: bool = Field(
        default=True,
        description="Whether to gaussianize the resulting exposures.",
    )
    gaussianize_maintain_zeros: bool = Field(
        default=False,
        description="Whether to maintain zeros when gaussianizing the exposures.",
    )

    @property
    def hierarchies(self) -> list[HierarchyType]:
        """The list of hierarchies for the exposure group.

        Returns
        -------
        list[HierarchyType]
            Empty list; tsbeta exposures have no hierarchies.
        """
        return []

    def effective_exposure_group_factors(
        self,
        menu: ExposureSettingsMenu,
    ) -> list[str]:
        """Get the effective factors.

        Parameters
        ----------
        menu : ExposureSettingsMenu
            The menu to get the total factors from (unused).

        Returns
        -------
        list[str]
            Empty list; factors are resolved at report time from uploaded data.
        """
        return []

    def normalize_group_settings(
        self,
        menu: ExposureSettingsMenu,
        universe_filter_dict: Mapping[str, CategoricalFilterSettings] | None,
    ) -> TSBetaExposureGroupSettings:
        """Normalize the exposure group settings.

        Parameters
        ----------
        menu : ExposureSettingsMenu
            The menu to get the total factors from (unused).
        universe_filter_dict : Mapping[str, CategoricalFilterSettings] | None
            The universe filter dictionary (unused).

        Returns
        -------
        TSBetaExposureGroupSettings
            Self, unchanged; tsbeta groups do not need normalization.
        """
        return self


class InteractionExposureGroupSettings(
    BaseExposureGroupSettings, frozen=True, extra="forbid"
):
    """The settings for an interaction exposure group.

    Interaction exposures are exposures that are a combination of two or more
    exposure groups. For example, we may want to create industry-specific style factors.
    """

    exposure_type: Literal["interaction"] = "interaction"
    exposure_groups: list[
        ContinuousExposureGroupSettings | CategoricalExposureGroupSettings
    ] = Field(
        [],
        description="The exposure groups to use for the Cartesian product.",
        min_length=2,
    )
    factor_group: str = Field(
        "",  # temporary value that does not pass validation
        description=(
            "By default, the name of the factor group will be a concatenation of the "
            "names of the exposure groups separated by the a colon."
        ),
        examples=["industry:style"],
    )

    @model_validator(mode="before")
    @classmethod
    def set_factor_group(cls, data: Any) -> Any:
        if isinstance(data, dict):
            if "factor_group" not in data or data["factor_group"] is None:
                factor_groups = [g.factor_group for g in data["exposure_groups"]]
                data["factor_group"] = ":".join(factor_groups)
        return data

    @field_validator("factor_group")
    @classmethod
    def check_valid_factor_group(cls, v: str) -> str:
        if v == "":
            raise ValueError("factor_group cannot be empty")
        return v

    @field_validator("exposure_groups", mode="after")
    @classmethod
    def validate_exposure_groups(
        cls, v: list[ExposureGroupSettings], info: ValidationInfo
    ) -> list[ExposureGroupSettings]:
        if isinstance(menu := info.context, ExposureSettingsMenu):
            for group_settings in v:
                group_settings.validate_exposure_group_settings(menu)
        return v

    @property
    def hierarchies(self) -> list[HierarchyType]:
        """The list of hierarchies for the exposure group.

        Returns
        -------
        list[HierarchyType]
            The list of hierarchies for the exposure group.
        """
        return [h for g in self.exposure_groups for h in g.hierarchies]

    def effective_exposure_group_factors(
        self,
        menu: ExposureSettingsMenu,
    ) -> list[str]:
        """Get the effective factors.

        Effective factors are the factors that are included and not excluded in the
        exposure group settings.

        Parameters
        ----------
        menu : ExposureSettingsMenu
            The menu to get the total factors from.

        Returns
        -------
        list[str]
            The list of effective factors.
        """
        return [
            f
            for g in self.exposure_groups
            for f in g.effective_exposure_group_factors(menu)
        ]

    def normalize_group_settings(
        self,
        menu: ExposureSettingsMenu,
        universe_filter_dict: Mapping[str, CategoricalFilterSettings] | None,
    ) -> InteractionExposureGroupSettings:
        """Normalize the exposure group settings.

        Normalized group settings have all hierarchies replaced by HierarchyGroups with
        two levels, the factor names and the leaf codes. All includes and excludes are
        resolved.

        This means that if the universe_filter_dict has a filter that trims down to just
        the North American region, and the ExposureSettings include factors from the
        same country hierarchy, then the factors will be trimmed down to just include
        the ones corresponding to the North American region.

        Parameters
        ----------
        menu : ExposureSettingsMenu
            The menu to get the total factors from.
        universe_filter_dict : Mapping[str, CategoricalFilterSettings] | None
            The universe filter dictionary to potentially further cut the factors down.
            The keys are the hierarchy names and the values are the filter settings.
            If None, then no filtering is performed.

        Returns
        -------
        InteractionExposureGroupSettings
            The normalized exposure group settings.
        """
        return self.model_copy(
            update={
                "exposure_groups": [
                    exposure_group.normalize_group_settings(menu, universe_filter_dict)
                    for exposure_group in self.exposure_groups
                ]
            },
            deep=True,
        )


ExposureGroupSettings = Annotated[
    ContinuousExposureGroupSettings
    | CategoricalExposureGroupSettings
    | TSBetaExposureGroupSettings
    | InteractionExposureGroupSettings,
    Field(discriminator="exposure_type"),
]


def _normalize_hierarchy(
    hierarchy: HierarchyLevel | HierarchyGroups,
    reference_hierarchy: Hierarchy,
    reference_labels: Mapping[str, str],
) -> HierarchyGroups:
    groupings: dict[str, list[str]] = {}
    if isinstance(hierarchy, HierarchyLevel):
        keys = settings_tools.flatten(
            settings_tools.trim_to_depth(reference_hierarchy, hierarchy.level),
            only_leaves=True,
        )
        for k in keys:
            sub_hierarchy = settings_tools.find_in_hierarchy(reference_hierarchy, k)
            if sub_hierarchy is not None:
                k = reference_labels.get(k, k)  # convert top level to label if possible
                groupings[k] = settings_tools.flatten(
                    sub_hierarchy,
                    only_leaves=True,
                )
    else:  # HierarchyGroups
        for k, v in hierarchy.groupings.items():
            groupings[k] = []
            # v is a Hierarchy, first map the codes to labels (with ~ prefix)
            v = settings_tools.labels_to_codes(v, reference_labels, prefix="~")
            # then process the excludes
            v = settings_tools.process_excludes(v, reference_hierarchy)
            # then flatten the hierarchy
            for code in v:
                sub_hierarchy = settings_tools.find_in_hierarchy(
                    reference_hierarchy, code
                )
                if sub_hierarchy is not None:
                    groupings[k].extend(
                        settings_tools.flatten(sub_hierarchy, only_leaves=True)
                    )
    return HierarchyGroups(name=hierarchy.name, groupings=groupings)


def _effective_exposure_group_factors(
    delegate: ContinuousExposureGroupSettings | CategoricalExposureGroupSettings,
    menu: ExposureSettingsMenu,
) -> list[str]:
    delegate.validate_exposure_group_settings(menu)
    hierarchy_name = delegate.hierarchy.name
    hierarchy_groups = _normalize_hierarchy(
        delegate.hierarchy,
        menu.hierarchies[hierarchy_name],
        menu.hierarchies_labels[hierarchy_name],
    )
    # filter based on include and exclude
    effective_groupings = settings_tools.effective_hierarchy(
        hierarchy_groups.groupings,
        delegate.include,
        delegate.exclude,
        menu.hierarchies_labels[hierarchy_name],
    )
    # just return the keys of effective_groupings
    return list(effective_groupings)


@overload
def _normalize_group_settings(
    delegate: ContinuousExposureGroupSettings,
    menu: ExposureSettingsMenu,
    universe_filter_dict: Mapping[str, CategoricalFilterSettings] | None,
) -> ContinuousExposureGroupSettings: ...


@overload
def _normalize_group_settings(
    delegate: CategoricalExposureGroupSettings,
    menu: ExposureSettingsMenu,
    universe_filter_dict: Mapping[str, CategoricalFilterSettings] | None,
) -> CategoricalExposureGroupSettings: ...


def _normalize_group_settings(
    delegate: ContinuousExposureGroupSettings | CategoricalExposureGroupSettings,
    menu: ExposureSettingsMenu,
    universe_filter_dict: Mapping[str, CategoricalFilterSettings] | None,
) -> ContinuousExposureGroupSettings | CategoricalExposureGroupSettings:
    # first normalize the hierarchy to Mapping[str, list[str]] with all leaf codes
    name = delegate.hierarchy.name
    hierarchy_groups = _normalize_hierarchy(
        delegate.hierarchy,
        menu.hierarchies[name],
        menu.hierarchies_labels[name],
    )
    # filter based on include and exclude
    new_groupings = settings_tools.effective_hierarchy(
        hierarchy_groups.groupings,
        delegate.include,
        delegate.exclude,
        menu.hierarchies_labels[name],
    )
    hierarchy_groups = hierarchy_groups.model_copy(
        update={"groupings": new_groupings}, deep=True
    )
    if universe_filter_dict:
        # filter the hierarchy groupings down if the universe has filters
        if name in universe_filter_dict:  # filter the hierarchy groupings down
            filter_settings = universe_filter_dict[name]
            effective_leaves = settings_tools.effective_leaves(
                menu.hierarchies[name],
                filter_settings.include,
                filter_settings.exclude,
                menu.hierarchies_labels[name],
            )
            new_groupings = {
                k: filtered_leaves
                for k, v in hierarchy_groups.groupings.items()
                if (filtered_leaves := [w for w in v if w in effective_leaves])
            }
            hierarchy_groups = hierarchy_groups.model_copy(
                update={"groupings": new_groupings}, deep=True
            )

    # normalized group settings should not have includes / excludes
    return delegate.model_copy(
        update={"hierarchy": hierarchy_groups, "include": "All", "exclude": []},
        deep=True,
    )


class ExposureSettingsMenu(SettingsMenu, frozen=True, extra="forbid"):
    """Contains the available settings that can be used to define exposures."""

    universe_settings_menu: UniverseSettingsMenu = Field(
        description="The universe settings menu to use to define the exposures.",
    )

    continuous_hierarchies: Mapping[str, Hierarchy] = Field(
        description="""
        A dictionary where the key is the name of the continuous hierarchy (e.g. 'style')
        and the value is a N-level nested dictionary structure of the continuous hierarchy
        codes.
        """,
    )

    continuous_hierarchies_labels: Mapping[str, Mapping[str, str]] = Field(
        description="""
        A dictionary where the key is the name of the continuous hierarchy and
        the value is a mapping from unique continuous code to a human readable name.
        """,
    )

    @property
    def calendar_settings_menu(self) -> CalendarSettingsMenu:
        """The calendar settings menu from the universe settings menu.

        Returns
        -------
        CalendarSettingsMenu
            The calendar settings menu from the universe settings menu.
        """
        return self.universe_settings_menu.calendar_settings_menu

    @property
    def categorical_hierarchies(self) -> Mapping[str, Hierarchy]:
        """
        The categorical hierarchies from the universe settings menu.

        Returns
        -------
        Mapping[str, Hierarchy]
            A dictionary where the key is the name of the categorical hierarchy (e.g. 'trbc')
            and the value is a N-level nested dictionary structure of the categorical hierarchy
            codes.
        """
        return self.universe_settings_menu.categorical_hierarchies

    @property
    def categorical_hierarchies_labels(self) -> Mapping[str, Mapping[str, str]]:
        """
        The categorical hierarchies labels from the universe settings menu.

        Returns
        -------
        Mapping[str, Mapping[str, str]]
            A dictionary where the key is the name of the categorical hierarchy and
            the value is a mapping from unique categorical code to a human readable name.
        """
        return self.universe_settings_menu.categorical_hierarchies_labels

    @property
    def hierarchies(self) -> Mapping[str, Hierarchy]:
        """The combined hierarchies from the continuous and categorical hierarchies.

        Returns
        -------
        Mapping[str, Hierarchy]
            A dictionary where the key is the name of the hierarchy and
            the value is the hierarchy.
        """
        return {**self.continuous_hierarchies, **self.categorical_hierarchies}

    @property
    def hierarchies_labels(self) -> Mapping[str, Mapping[str, str]]:
        """The combined hierarchies labels from the continuous and categorical hierarchies.

        Returns
        -------
        Mapping[str, Mapping[str, str]]
            A dictionary where the key is the name of the hierarchy and
            the value is a mapping from unique code to a human readable name.
        """
        return {
            **self.continuous_hierarchies_labels,
            **self.categorical_hierarchies_labels,
        }

    def describe(self) -> str:
        """Describe the exposure settings menu.

        Returns
        -------
        str
            The description of the exposure settings menu.
        """
        # only describe continuous hierarchies for now
        hierarchies = self.continuous_hierarchies
        labels = self.continuous_hierarchies_labels

        hierarchies_str = json.dumps(
            {
                k: settings_tools.codes_to_labels(hierarchies[k], labels[k])
                for k in hierarchies
            },
            indent=2,
        )
        description = [
            "Hierarchies:",
            hierarchies_str,
        ]

        return os.linesep.join(description)

    @field_validator("continuous_hierarchies")
    @classmethod
    def check_unique_hierarchy(
        cls, v: Mapping[str, Hierarchy]
    ) -> Mapping[str, Hierarchy]:
        return settings_tools.check_unique_hierarchy(v)

    @field_validator("continuous_hierarchies")
    @classmethod
    def check_no_empty_branches(
        cls, v: Mapping[str, Hierarchy]
    ) -> Mapping[str, Hierarchy]:
        for hierarchy_name, hierarchy in v.items():
            if not hierarchy:  # different error message at root level
                raise ValueError(f"Hierarchy '{hierarchy_name}' cannot be empty")
            settings_tools.assert_no_empty_branches(hierarchy, hierarchy_name)
        return v

    @model_validator(mode="after")
    def check_all_codes_have_labels(self) -> ExposureSettingsMenu:
        if errors := settings_tools.check_all_codes_have_labels(
            self.hierarchies, self.hierarchies_labels
        ):
            raise ValueError(os.linesep.join(errors))
        else:
            return self

    def hierarchy_df(self, hierarchy: str) -> pl.DataFrame:
        """Return a dataframe of the given hierarchy.

        Parameters
        ----------
        hierarchy: str
            the name of the hierarchy to return.

        Returns
        -------
        pl.DataFrame
            The wide DataFrame representation of the hierarchy. The columns are:
            - level_1: The code of the root level.
            - level_1_label: The label of the root level.
            - level_2: The code of the second level.
            - level_2_label: The label of the second level.
            - ...
            - level_n: The code of the n-th level.
            - level_n_label: The label of the n-th level.
        """
        return settings_tools.hierarchy_df_to_wide(
            self.hierarchies[hierarchy], self.hierarchies_labels[hierarchy]
        )


class ExposureSettings(Settings[ExposureSettingsMenu]):
    """Defines exposures as hierarchy of selected styles and substyles."""

    exposures: Annotated[
        list[ExposureGroupSettings],
        BeforeValidator(_ensure_list),
        Field(
            description="The exposures to build the factor risk model on.",
        ),
    ] = []

    @property
    def menu_type(self) -> type[ExposureSettingsMenu]:  # noqa: D102
        return ExposureSettingsMenu

    def describe(self, menu: ExposureSettingsMenu) -> str:
        """Describe the exposure settings.

        Parameters
        ----------
        menu : ExposureSettingsMenu
            The menu to get context information from.

        Returns
        -------
        str
            The description of the exposure settings.
        """
        # only describe continuous hierarchies for now
        hierarchies = menu.continuous_hierarchies
        labels = menu.continuous_hierarchies_labels

        self.validate_settings(menu)  # TODO: needed?
        self_normalized = self.normalize(None, menu)
        description = []
        for group_settings in self_normalized.exposures:
            if group_settings.exposure_type == "continuous":  # only continuous
                hierarchy_name = group_settings.hierarchy.name
                hierarchy = settings_tools.codes_to_labels(
                    hierarchies[hierarchy_name], labels[hierarchy_name]
                )
                description.extend(
                    [
                        f"Hierarchy ({hierarchy_name}):",
                        json.dumps(hierarchy, indent=2),
                    ]
                )
        return os.linesep.join(description)

    @field_validator("exposures", mode="after")
    @classmethod
    def validate_exposure_groups(
        cls, exposures: list[ExposureGroupSettings], info: ValidationInfo
    ) -> list[ExposureGroupSettings]:
        if isinstance(menu := info.context, ExposureSettingsMenu):
            factor_dict: dict[str, list[str]] = {}  # collect factors per factor_group
            for group_settings in exposures:
                group_settings.validate_exposure_group_settings(menu)
                factor_dict.setdefault(group_settings.factor_group, []).extend(
                    group_settings.effective_exposure_group_factors(menu)
                )
            # check that output factor_group-factor combinations are unique
            settings_tools.check_unique_hierarchy(factor_dict)

        return exposures

    def normalize(
        self,
        universe_settings: UniverseSettings | None,
        exposure_settings_menu: ExposureSettingsMenu,
    ) -> ExposureSettings:
        """Normalize the given exposure settings.

        Normalize the given exposure settings by converting all exposure hierarchies
        to a HierarchyGroups object, and then filtering the hierarchy groupings based
        on the include and exclude statements in the exposure settings, and possibly
        the universe filters.

        This means that if the UniverseSettings have a filter that trims down to just
        the North American region, and the ExposureSettings include factors from the
        same country hierarchy, then the factors will be trimmed down to just include
        the ones corresponding to the North American region.

        Parameters
        ----------
        universe_settings: UniverseSettings | None
            the universe settings to use for normalization. If None, then no
            normalization is performed in relation to the filters in the universe.
        exposure_settings_menu: ExposureSettingsMenu
            The menu to get the total factors from.

        Returns
        -------
        ExposureSettings
            A new exposure settings object with all exposure hierarchies converted to
            HierarchyGroups objects, and the hierarchy groupings filtered based on the
            include and exclude statements in the exposure settings, and possibly the
            universe filters. This normalized object is also validated.
        """
        if universe_settings:
            universe_filter_dict = {
                filter_settings.hierarchy: filter_settings  # hierarchy is unique
                for filter_settings in universe_settings.categorical_filters
            }
        else:
            universe_filter_dict = None

        self_normalized = self.model_copy(
            update={
                "exposures": [
                    group_settings.normalize_group_settings(
                        exposure_settings_menu, universe_filter_dict
                    )
                    for group_settings in self.exposures
                ]
            },
            deep=True,
        )

        # we may have produced an invalid settings object with the copy/update calls
        # so we need to validate again (validate_settings does pydantic validation)
        self_normalized.validate_settings(exposure_settings_menu)

        return self_normalized



import abc
from typing import Generic, ParamSpec, TypeVar

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.settings import ReportSettings
from bayesline.api._src.tasks import AsyncTask, Task

T = TypeVar("T", bound=TypedReportAccessorApi, covariant=True)
AT = TypeVar("AT", bound=AsyncTypedReportAccessorApi, covariant=True)
S = TypeVar("S", bound=ReportSettings)
P = ParamSpec("P")


class ReportApi(abc.ABC, Generic[P, T, S]):
    """
    A base interface for report APIs.

    It is meant to be extended by several more concrete interfaces which
    narrow down the set of possible args and kwargs to a more specific set,
    e.g. for the subset of possible reports that require a start and end date.
    """

    @property
    @abc.abstractmethod
    def settings(self) -> S:
        """
        The settings used to create this report.

        Returns
        -------
        S
            The settings used to create this report.
        """

    @abc.abstractmethod
    def get_maybe_cached(self, *args: P.args, **kwargs: P.kwargs) -> T | None:
        """
        Get the cached report accessor API if it is available.

        Parameters
        ----------
        *args: P.args
            The arguments to the report calculation.
        **kwargs: P.kwargs
            The keyword arguments to the report calculation.

        Returns
        -------
        T | None
            The cached report accessor API, or None if the report is not cached.
        """

    @abc.abstractmethod
    def calculate(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """
        Triggers the calculation of the report and returns an accessor API.

        Parameters
        ----------
        *args: P.args
            The arguments to the report calculation.
        **kwargs: P.kwargs
            The keyword arguments to the report calculation.

        Returns
        -------
        T
            The accessor API for the report result.
        """

    @abc.abstractmethod
    def calculate_as_task(  # noqa: D102
        self, *args: P.args, **kwargs: P.kwargs
    ) -> Task[T]: ...

    @abc.abstractmethod
    def to_mermaid(self) -> str:
        """
        Get a Mermaid diagram of the report execution graph.

        Returns
        -------
        str
            The Mermaid diagram of the report.
        """


@docstrings_from_sync
class AsyncReportApi(abc.ABC, Generic[P, AT, S]):
    @property
    @abc.abstractmethod
    def settings(self) -> S: ...  # noqa: D102

    @abc.abstractmethod
    async def get_maybe_cached(  # noqa: D102
        self, *args: P.args, **kwargs: P.kwargs
    ) -> AT | None: ...

    @abc.abstractmethod
    async def calculate(  # noqa: D102
        self, *args: P.args, **kwargs: P.kwargs
    ) -> AT: ...

    @abc.abstractmethod
    async def calculate_as_task(  # noqa: D102
        self, *args: P.args, **kwargs: P.kwargs
    ) -> AsyncTask[AT]: ...

    @abc.abstractmethod
    async def to_mermaid(self) -> str: ...  # noqa: D102



import os
from collections.abc import Sequence
from typing import Literal, Mapping

from pydantic import (
    Field,
    NonNegativeFloat,
    NonPositiveFloat,
    ValidationInfo,
    field_validator,
)

from bayesline.api._src.equity.universe_settings import UniverseSettings
from bayesline.api._src.registry import Settings, SettingsMenu

WeightingScheme = Literal["SqrtCap", "InvIdioVar"]


class ModelConstructionSettingsMenu(SettingsMenu, frozen=True, extra="forbid"):
    """Defines available modelconstruction settings to build a factor risk model."""

    weights: list[WeightingScheme] = Field(
        description="""
        The available regression weights that can be used for the factor risk model.
        """,
    )

    def describe(self) -> str:
        """Describe the available settings.

        Returns
        -------
        str
            A description of the available settings.
        """
        lines = [f"Weights: {', '.join(self.weights)}"]
        return os.linesep.join(lines)


class ModelConstructionSettings(Settings[ModelConstructionSettingsMenu]):
    """Defines settings to build a factor risk model."""

    currency: str = Field(
        description="The currency of the factor risk model.",
        default="USD",
        examples=["USD", "EUR"],
    )
    weights: WeightingScheme = Field(
        description="The regression weights used for the factor risk model.",
        default="SqrtCap",
        examples=["SqrtCap", "InvIdioVar"],
    )
    estimation_universe: str | int | UniverseSettings | None = Field(
        None,
        description="The universe settings to use for the estimation universe.",
    )
    return_clip_bounds: tuple[NonPositiveFloat | None, NonNegativeFloat | None] = Field(
        description="The bounds for the return clipping.",
        default=(-0.1, 0.1),
        examples=[(-0.1, 0.1), (None, None)],
    )
    thin_category_shrinkage: Mapping[str, NonNegativeFloat] = Field(
        default_factory=dict,
        description=(
            "The ridge-shrinkage penalty for categorical factors with the than 10 "
            "assets with strictly positive exposure. Interpolation is used to scale "
            "the shrinkage strength (for zero assets) to 0.0 (for 10 assets). The keys "
            "are the (categorical) factor groups and the values are the shrinkage "
            "penalty."
        ),
    )
    thin_category_shrinkage_overrides: Mapping[tuple[str, str], NonNegativeFloat] = (
        Field(
            default_factory=dict,
            description=(
                "The shrinkage strength override for the factor risk model. The keys are "
                "the tuples of (categorical) factor groups and factor names and the values "
                "are the shrinkage penalty overrides."
            ),
        )
    )
    zero_sum_constraints: Mapping[
        str, Literal["none", "equal_weights", "mcap_weighted"]
    ] = Field(
        default_factory=dict,
        description=(
            "Whether to apply a zero-sum constraint to the categorical exposures. If the "
            "category exposures are exhaustive and sum to one, then implicitly a dummy "
            "variable trap is present. This can be avoided by creating a constraint on "
            "the factor returns. This means the interpretation of the categorical "
            "factor returns is 'in excess of' the market. If 'none', no constraint is "
            "applied (for example if we do not have a market factor). If "
            "'equal_weights' or 'mcap_weighted', then the categorical factor returns "
            "are constrained to sum to zero, either market-cap weighted or not. For "
            "all categorical factor groups, a value must be provided."
        ),
        examples=[
            {"industry": ["none"]},
            {"industry": ["equal_weights"], "country": ["mcap_weighted"]},
        ],
    )
    known_factor_map: Mapping[tuple[str, str], str] = Field(
        default_factory=dict,
        description=(
            "A mapping from tuple of factor groups and factor names (labels) to known "
            "series. These factors will have fixed returns derived from the series."
        ),
    )
    fx_convert_returns: bool = Field(
        default=True,
        description=(
            "Whether to convert the asset returns to the currency of the factor model."
        ),
    )

    def get_references(self) -> Sequence[str | int]:
        """
        Get references for this settings object.

        Returns
        -------
        Sequence[str | int]
            A sequence of references (strings or integers) for this settings object.
        """
        if self.estimation_universe is not None:
            if isinstance(self.estimation_universe, UniverseSettings):
                return self.estimation_universe.get_references()
            else:
                return [self.estimation_universe]
        return []

    @property
    def menu_type(self) -> type[ModelConstructionSettingsMenu]:  # noqa: D102
        return ModelConstructionSettingsMenu

    @field_validator("weights")
    @classmethod
    def check_valid_weights(
        cls, weights: WeightingScheme, info: ValidationInfo
    ) -> WeightingScheme:
        if isinstance(menu := info.context, ModelConstructionSettingsMenu):
            if weights not in menu.weights:
                raise ValueError(f"Invalid weights: {weights}")
        return weights

    def describe(self, menu: ModelConstructionSettingsMenu) -> str:
        """Describe the available settings.

        Parameters
        ----------
        menu : ModelConstructionSettingsMenu
            The menu to get context information from.

        Returns
        -------
        str
            The description of the model construction settings.
        """
        del menu  # not used
        lines = [
            f"Currency: {self.currency}",
            f"Weights: {self.weights}",
            f"Return Clip Bounds: {self.return_clip_bounds}",
        ]
        return os.linesep.join(lines)



import datetime as dt
import enum
from collections.abc import Sequence

from pydantic import BaseModel, ConfigDict, Field


class PermissionLevel(str, enum.Enum):
    """Permission levels for resource access."""

    READ = "read"
    WRITE = "write"
    DELETE = "delete"
    ADMIN = "admin"  # Full control including permission management


class ResourceType(str, enum.Enum):
    """Types of resources that can be protected by ACL."""

    SETTING = "setting"
    # Future resource types can be added here:
    # DATASET, MODEL, REPORT, etc.


class PrincipalType(str, enum.Enum):
    """Types of principals that can be granted permissions."""

    USER = "user"
    GROUP = "group"
    ROLE = "role"


class RoleType(str, enum.Enum):
    """Built-in system roles with predefined permissions."""

    ADMIN = "admin"  # Can access and modify all resources
    USER = "user"  # Standard user with no special privileges


# strip-hints: off


class ACLEntry(BaseModel):
    """An individual ACL entry granting specific permissions to a principal for a resource."""

    model_config = ConfigDict(frozen=True)

    id: int
    resource_type: ResourceType
    resource_id: int  # The ID of the specific resource (e.g., settings.id)
    principal_type: PrincipalType
    principal_id: str  # User ID, group name, or role name
    permission_level: PermissionLevel
    granted_by: str  # User ID of who granted this permission
    granted_at: dt.datetime
    expires_at: dt.datetime | None = None  # Optional expiration
    created_at: dt.datetime
    updated_at: dt.datetime


class CreateACLEntry(BaseModel):
    """Request model for creating a new ACL entry."""

    model_config = ConfigDict(frozen=True)

    resource_type: ResourceType
    resource_id: int
    principal_type: PrincipalType
    principal_id: str
    permission_level: PermissionLevel
    granted_by: str
    expires_at: dt.datetime | None = None


class UpdateACLEntry(BaseModel):
    """Request model for updating an existing ACL entry."""

    model_config = ConfigDict(frozen=True)

    id: int
    permission_level: PermissionLevel | None = None
    expires_at: dt.datetime | None = None
    updated_by: str


class UserGroup(BaseModel):
    """A group of users for simplified permission management."""

    model_config = ConfigDict(frozen=True)

    id: int
    name: str = Field(pattern=r"^[A-Za-z0-9_-]+$")
    description: str
    created_by: str
    created_at: dt.datetime
    updated_at: dt.datetime


class CreateUserGroup(BaseModel):
    """Request model for creating a new user group."""

    model_config = ConfigDict(frozen=True)

    name: str = Field(pattern=r"^[A-Za-z0-9_-]+$")
    description: str
    created_by: str


class UpdateUserGroup(BaseModel):
    """Request model for updating an existing user group."""

    model_config = ConfigDict(frozen=True)

    id: int
    description: str | None = None
    updated_by: str


class GroupMembership(BaseModel):
    """Represents a user's membership in a group."""

    model_config = ConfigDict(frozen=True)

    id: int
    group_id: int
    user_id: str
    added_by: str
    added_at: dt.datetime


class CreateGroupMembership(BaseModel):
    """Request model for adding a user to a group."""

    model_config = ConfigDict(frozen=True)

    group_id: int
    user_id: str
    added_by: str


class UserRole(BaseModel):
    """Represents a user's system role assignment."""

    model_config = ConfigDict(frozen=True)

    id: int
    user_id: str
    role_type: RoleType
    assigned_by: str
    assigned_at: dt.datetime
    expires_at: dt.datetime | None = None


class CreateUserRole(BaseModel):
    """Request model for assigning a role to a user."""

    model_config = ConfigDict(frozen=True)

    user_id: str
    role_type: RoleType
    assigned_by: str
    expires_at: dt.datetime | None = None


class ResourcePermissionSummary(BaseModel):
    """Summary of a user's permissions for a specific resource."""

    model_config = ConfigDict(frozen=True)

    resource_type: ResourceType
    resource_id: int
    resource_name: str | None = None  # Human-readable name if available
    resource_owner: str | None = None  # Owner of the resource
    effective_permission: PermissionLevel
    permission_sources: Sequence[str]  # How permission was granted


class UserPermissionsSummary(BaseModel):
    """Complete summary of a user's permissions across all resources."""

    model_config = ConfigDict(frozen=True)

    user_id: str
    roles: Sequence[RoleType]
    groups: Sequence[str]  # Group names
    resources: Sequence[ResourcePermissionSummary]


# strip-hints: on


class ACLError(Exception):
    """Base exception for ACL-related errors."""

    pass


class PermissionDeniedError(ACLError):
    """Raised when a user attempts an action they don't have permission for."""

    pass


class ResourceNotFoundError(ACLError):
    """Raised when a resource referenced in ACL operations doesn't exist."""

    pass


class PrincipalNotFoundError(ACLError):
    """Raised when a principal (user/group/role) referenced in ACL operations doesn't exist."""

    pass


class ACLValidationError(ACLError):
    """Raised when ACL data fails validation."""

    pass


class PermissionGrantError(ACLError):
    """Raised when granting a permission fails."""

    pass


class GroupCreationError(ACLError):
    """Raised when creating a group fails."""

    pass


class GroupUpdateError(ACLError):
    """Raised when updating a group fails."""

    pass


class GroupDeletionError(ACLError):
    """Raised when deleting a group fails."""

    pass


class GroupMembershipError(ACLError):
    """Raised when adding or removing users from groups fails."""

    pass


class RoleAssignmentError(ACLError):
    """Raised when assigning or revoking roles fails."""

    pass



import abc
from typing import Any, Literal, Self

import polars as pl
from pydantic import Field, model_validator

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.calendar_settings import CalendarSettings
from bayesline.api._src.equity.portfoliohierarchy_settings import (
    PortfolioHierarchySettings,
)
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    ReportAccessorApi,
)
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.api import AsyncReportApi, ReportApi
from bayesline.api._src.equity.report.scenarios import (
    PortfolioHoldingsScenario,
)
from bayesline.api._src.equity.report.settings import ReportSettings
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike, DNFFilterExpressions


class PortfolioReturnsReportSettings(ReportSettings):
    """Settings for a portfolio returns report."""

    report_type: Literal["portfolio_returns"] = Field(
        default="portfolio_returns", description="Report type discriminator"
    )
    portfolio_hierarchy_settings: PortfolioHierarchySettings = Field()
    calendar: CalendarSettings = Field(default_factory=CalendarSettings)
    currency: str | None = Field(None)
    returns_currency: str = Field("USD")
    rescale_bench: bool = Field(True)
    normalize_holdings: bool = Field(True)
    return_aggregation_type: Literal["arithmetic", "geometric"] = Field("geometric")

    @model_validator(mode="after")
    def validate_geometric_requires_normalized(self) -> Self:
        if self.return_aggregation_type == "geometric" and not self.normalize_holdings:
            raise ValueError(
                "Geometric compounding requires normalize_holdings=True. "
                "PnL values cannot be geometrically compounded."
            )
        return self


class PortfolioReturnsReportAccessor(TypedReportAccessorApi):
    """Specific accessor for a portfolio returns report."""

    scenario_types = [PortfolioHoldingsScenario]

    @abc.abstractmethod
    def get_portolio_returns(
        self,
        portfolio_names: list[str] | None = None,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        aggregation_type: Literal["D", "M", "Y"] | None = "D",
        value_cols: list[str] | None = None,
    ) -> pl.DataFrame:
        """
        Get the portfolio returns for the portfolio names between start and end dates.

        Parameters
        ----------
        portfolio_names: list[str] | None, default=None
            The names of the portfolios to calculate the report for. If None, all
            portfolios will be calculated.
        start_date: DateLike | None, default=None
            The start date of the report.
        end_date: DateLike | None, default=None
            The end date of the report.
        aggregation_type: Literal["D", "M", "Y"] | None, default="D"
            The aggregation type to use for the returns.
            If None, dates will be completely collapsed and the returns for the entire
            period will be returned.
        value_cols: list[str] | None, default=None
            The value columns to return. If None, all value columns will be returned.

        Returns
        -------
        pl.DataFrame
            A dataframe with the portfolio returns.
        """

    @abc.abstractmethod
    def get_asset_returns(
        self,
        portfolio_id: str,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        aggregation_type: Literal["D", "M", "Y"] | None = "D",
        value_cols: list[str] | None = None,
    ) -> pl.DataFrame:
        """
        Get the asset level returns for the portfolio id between start and end dates.

        Parameters
        ----------
        portfolio_id: str
            The id of the portfolio to calculate the report for.
        start_date: DateLike | None, default=None
            The start date of the report.
        end_date: DateLike | None, default=None
            The end date of the report.
        aggregation_type: Literal["D", "M", "Y"] | None, default="D"
            The aggregation type to use for the returns.
            If None, dates will be completely collapsed and the returns for the entire
            period will be returned.
        value_cols: list[str] | None, default=None
            The value columns to return. If None, all value columns will be returned.

        Returns
        -------
        pl.DataFrame
            A dataframe with the asset level returns.
        """


@docstrings_from_sync
class AsyncPortfolioReturnsReportAccessor(AsyncTypedReportAccessorApi):
    scenario_types = [PortfolioHoldingsScenario]

    @abc.abstractmethod
    async def get_portolio_returns(  # noqa: D102
        self,
        portfolio_names: list[str] | None = None,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        aggregation_type: Literal["D", "M", "Y"] | None = "D",
        value_cols: list[str] | None = None,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def get_asset_returns(  # noqa: D102
        self,
        portfolio_id: str,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        aggregation_type: Literal["D", "M", "Y"] | None = "D",
        value_cols: list[str] | None = None,
    ) -> pl.DataFrame: ...


class PortfolioReturnsReportApi(
    ReportApi[
        [list[str] | None, DateLike | None, DateLike | None],
        PortfolioReturnsReportAccessor,
        PortfolioReturnsReportSettings,
    ]
):
    """API for a portfolio returns report."""

    @abc.abstractmethod
    def get_maybe_cached(
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> PortfolioReturnsReportAccessor | None:
        """
        Get the cached portfolio returns report accessor API if it is available.

        Parameters
        ----------
        portfolio_names: list[str] | None
            The names of the portfolios to calculate the report for. If None, all
            portfolios will be calculated.
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        PortfolioReturnsReportAccessor | None
            The cached portfolio returns report accessor API, or None if the report is
            not cached.
        """

    @abc.abstractmethod
    def calculate(
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> PortfolioReturnsReportAccessor:
        """
        Calculate the portfolio returns report.

        Parameters
        ----------
        portfolio_names: list[str] | None
            The names of the portfolios to calculate the report for. If None, all
            portfolios will be calculated.
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        PortfolioReturnsReportAccessor
            The portfolio returns report accessor.
        """

    @abc.abstractmethod
    def calculate_as_task(
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> Task[PortfolioReturnsReportAccessor]:
        """
        Calculate the portfolio returns report as a task.

        Parameters
        ----------
        portfolio_names: list[str] | None
            The names of the portfolios to calculate the report for. If None, all
            portfolios will be calculated.
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        Task[PortfolioReturnsReportAccessor]
            The task for the portfolio returns report accessor.
        """


@docstrings_from_sync
class AsyncPortfolioReturnsReportApi(
    AsyncReportApi[
        [list[str] | None, DateLike | None, DateLike | None],
        AsyncPortfolioReturnsReportAccessor,
        PortfolioReturnsReportSettings,
    ]
):

    @abc.abstractmethod
    async def get_maybe_cached(  # noqa: D102
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncPortfolioReturnsReportAccessor | None: ...

    @abc.abstractmethod
    async def calculate(  # noqa: D102
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncPortfolioReturnsReportAccessor: ...

    @abc.abstractmethod
    async def calculate_as_task(  # noqa: D102
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncTask[AsyncPortfolioReturnsReportAccessor]: ...


def _get_portfolio_returns_filters(
    start_date: DateLike | None,
    end_date: DateLike | None,
    portfolio_names: list[str] | None,
) -> DNFFilterExpressions:
    filters: DNFFilterExpressions = []
    if start_date is not None:
        filters.append(("date", ">=", start_date))
    if end_date is not None:
        filters.append(("date", "<=", end_date))
    if portfolio_names is not None:
        filters.append(("portfolio_id", "in", portfolio_names))
    return filters


def _get_portfolio_returns_expand(
    aggregation_type: Literal["D", "M", "Y"] | None = "D",
) -> tuple[str, ...]:
    expand = ["portfolio_id"]
    if aggregation_type == "D":
        expand.extend(["year", "month", "date"])
    elif aggregation_type == "M":
        expand.extend(["year", "month"])
    elif aggregation_type == "Y":
        expand.extend(["year"])
    return tuple(expand)


class PortfolioReturnsReportAccessorImpl(PortfolioReturnsReportAccessor):  # noqa: D101

    def __init__(self, accessor: ReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> ReportAccessorApi:  # noqa: D102
        return self._accessor

    def get_portolio_returns(  # noqa: D102
        self,
        portfolio_names: list[str] | None = None,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        aggregation_type: Literal["D", "M", "Y"] | None = "D",
        value_cols: list[str] | None = None,
    ) -> pl.DataFrame:
        available_value_cols = self._accessor.metric_cols
        if value_cols is not None:
            if set(value_cols) > set(available_value_cols):
                raise ValueError(
                    f"Unknown value_cols: "
                    f"{set(value_cols) - set(available_value_cols)}. "
                    f"Available: {available_value_cols}."
                )
        else:
            value_cols = available_value_cols
        return self._accessor.get_data(
            path=list(),
            expand=_get_portfolio_returns_expand(aggregation_type),
            filters=_get_portfolio_returns_filters(
                start_date, end_date, portfolio_names
            ),
            value_cols=tuple(value_cols),
        )

    def get_asset_returns(  # noqa: D102
        self,
        portfolio_id: str,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        aggregation_type: Literal["D", "M", "Y"] | None = "D",
        value_cols: list[str] | None = None,
    ) -> pl.DataFrame:
        available_value_cols = self._accessor.metric_cols
        if value_cols is not None:
            if set(value_cols) > set(available_value_cols):
                raise ValueError(
                    f"Unknown value_cols: "
                    f"{set(value_cols) - set(available_value_cols)}. "
                    f"Available: {available_value_cols}."
                )
        else:
            value_cols = available_value_cols
        return self._accessor.get_data(
            path=[("portfolio_id", portfolio_id)],
            expand=tuple(
                list(
                    [
                        p
                        for p in _get_portfolio_returns_expand(aggregation_type)
                        if p != "portfolio_id"
                    ]
                    + [
                        "input_asset_id",
                        "input_asset_id_type",
                        "asset_id",
                        "asset_name",
                    ]
                )
            ),
            filters=_get_portfolio_returns_filters(
                start_date, end_date, [portfolio_id]
            ),
            value_cols=tuple(value_cols),
        )


class AsyncPortfolioReturnsReportAccessorImpl(  # noqa: D101
    AsyncPortfolioReturnsReportAccessor
):

    def __init__(self, accessor: AsyncReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> AsyncReportAccessorApi:  # noqa: D102
        return self._accessor

    async def get_portolio_returns(  # noqa: D102
        self,
        portfolio_names: list[str] | None = None,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        aggregation_type: Literal["D", "M", "Y"] | None = "D",
        value_cols: list[str] | None = None,
    ) -> pl.DataFrame:
        available_value_cols = self._accessor.metric_cols
        if value_cols is not None:
            if set(value_cols) > set(available_value_cols):
                raise ValueError(
                    f"Unknown value_cols: "
                    f"{set(value_cols) - set(available_value_cols)}. "
                    f"Available: {available_value_cols}."
                )
        else:
            value_cols = available_value_cols
        return await self._accessor.get_data(
            path=[],
            expand=_get_portfolio_returns_expand(aggregation_type),
            filters=_get_portfolio_returns_filters(
                start_date, end_date, portfolio_names
            ),
            value_cols=tuple(value_cols) if value_cols is not None else (),
        )

    async def get_asset_returns(  # noqa: D102
        self,
        portfolio_id: str,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        aggregation_type: Literal["D", "M", "Y"] | None = "D",
        value_cols: list[str] | None = None,
    ) -> pl.DataFrame:
        available_value_cols = self._accessor.metric_cols
        if value_cols is not None:
            if set(value_cols) > set(available_value_cols):
                raise ValueError(
                    f"Unknown value_cols: "
                    f"{set(value_cols) - set(available_value_cols)}. "
                    f"Available: {available_value_cols}."
                )
        else:
            value_cols = available_value_cols
        return await self._accessor.get_data(
            path=[("portfolio_id", portfolio_id)],
            expand=tuple(
                list(
                    [
                        p
                        for p in _get_portfolio_returns_expand(aggregation_type)
                        if p != "portfolio_id"
                    ]
                    + [
                        "input_asset_id",
                        "input_asset_id_type",
                        "asset_id",
                        "asset_name",
                    ]
                )
            ),
            filters=_get_portfolio_returns_filters(
                start_date, end_date, [portfolio_id]
            ),
            value_cols=tuple(value_cols),
        )



import abc
import datetime as dt
from pathlib import Path
from typing import Literal

import polars as pl
from pydantic import BaseModel

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DNFFilterExpressions


class UploadDataTypeFreshness(BaseModel, frozen=True, extra="forbid"):
    """Freshness summary for an upload data type.

    Parameters
    ----------
    data_type : str
        The name of the upload data type (e.g., "portfolios", "exposures").
    last_committed_at : dt.datetime | None
        The most recent commit timestamp across all datasets of this data type.
        None if no datasets exist or no commits have been made.
    """

    data_type: str
    last_committed_at: dt.datetime | None = None


class UploadParserResult(BaseModel, frozen=True, extra="forbid"):
    """Result of a parser operation.

    Parameters
    ----------
    parser : str | None
        The name of the parser that was used.
    success : bool
        Whether the parsing was successful.
    messages : list[str]
        List of messages from the parsing operation.
    """

    parser: str | None = None  # TODO remove None
    success: bool
    messages: list[str]


class MultiParserResult(BaseModel, frozen=True, extra="forbid"):
    """Result of multiple parser operations.

    Parameters
    ----------
    results : list[UploadParserResult]
        List of parser results.
    """

    results: list[UploadParserResult]

    @property
    def success(self) -> bool:
        """Check if any parser was successful."""
        return any(result.success for result in self.results)

    @property
    def success_parser(self) -> UploadParserResult | None:
        """Get the first successful parser result."""
        return next((result for result in self.results if result.success), None)


class UploadStagingResult(BaseModel, frozen=True, extra="forbid"):
    """Result of a staging operation.

    Parameters
    ----------
    name : str
        The name of the staged item.
    timestamp : dt.datetime
        When the staging occurred.
    success : bool
        Whether the staging was successful.
    results : list[UploadParserResult]
        Results from the parsing operation.
    """

    name: str
    timestamp: dt.datetime
    success: bool
    results: list[UploadParserResult]

    @property
    def success_parser(self) -> UploadParserResult | None:
        """Get the first successful parser result."""
        return next((result for result in self.results if result.success), None)


class UploadCommitResult(BaseModel, frozen=True, extra="forbid"):
    """Result of a commit operation.

    Parameters
    ----------
    version : int
        The version number of the committed data.
    committed_names : list[str]
        Names of the committed items.
    """

    version: int
    committed_names: list[str]


class UploadParserApi(abc.ABC):
    """Abstract base class for upload parsers."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Get the name of the parser.

        Returns
        -------
        str
            The name of the parser.
        """
        ...

    @abc.abstractmethod
    def output_schema(self) -> dict[str, pl.DataType]:
        """Get the output schema that this parser produces.

        Returns
        -------
        dict[str, pl.DataType]
            The schema that this parser produces.
        """
        ...

    @abc.abstractmethod
    def get_examples(self) -> list[pl.DataFrame]:
        """Get example dataframes for testing the parser.

        Returns
        -------
        list[pl.DataFrame]
            A list of example portfolios that can be used to test the parser.
        """
        ...

    @abc.abstractmethod
    def can_handle(
        self, raw_df: pl.DataFrame, *, name: str | None = None
    ) -> UploadParserResult:
        """Check if this parser can handle the given dataframe.

        Parameters
        ----------
        raw_df : pl.DataFrame
            The dataframe to check if the parser can handle.
        name : str | None, default=None
            The name of the dataframe (which could be the filename).
            Some parsers might extract information (such as a date) from the name.

        Returns
        -------
        UploadParserResult
            Result indicating whether the parser can handle the dataframe.
        """
        ...

    @abc.abstractmethod
    def parse(
        self, raw_df: pl.DataFrame, *, name: str | None = None
    ) -> tuple[pl.DataFrame, UploadParserResult]:
        """Parse the given dataframe.

        Parameters
        ----------
        raw_df : pl.DataFrame
            The dataframe to parse.
        name : str | None, default=None
            The name of the dataframe (which could be the filename).
            Some parsers might extract information (such as a date) from the name.

        Returns
        -------
        tuple[pl.DataFrame, UploadParserResult]
            A tuple of the parsed pl.DataFrame and a UploadParserResult.
            If the parser was unsuccessful the parsed dataframe will instead be
            a (possibly empty) error dataframe.
        """
        ...


@docstrings_from_sync
class AsyncUploadParserApi(abc.ABC):

    @property
    @abc.abstractmethod
    def name(self) -> str: ...  # noqa: D102

    @abc.abstractmethod
    async def output_schema(self) -> dict[str, pl.DataType]: ...  # noqa: D102

    @abc.abstractmethod
    async def get_examples(self) -> list[pl.DataFrame]: ...  # noqa: D102

    @abc.abstractmethod
    async def can_handle(  # noqa: D102
        self, raw_df: pl.DataFrame, *, name: str | None = None
    ) -> UploadParserResult: ...

    @abc.abstractmethod
    async def parse(  # noqa: D102
        self, raw_df: pl.DataFrame, *, name: str | None = None
    ) -> tuple[pl.DataFrame, UploadParserResult]: ...


class UploadError(Exception):
    """Exception raised for upload-related errors."""

    pass


class UploaderApi(abc.ABC):
    """Provide functionality to parse, stage and upload data to a versioned storage.

    Parsing:
    This uploader contains a set of parsers that can be used to parse raw dataframes
    into their standardized format.

    Staging:
    The staging functionality allows to stage multiple dataframes for upload.
    This is done by providing a name and a dataframe for each staged dataframe
    and is not affecting the 'committed' data until the staged dataframes are committed.
    The staging area is unversioned.

    Commit:
    Allows to commit the staged dataframes to the versioned storage. All committed
    dataframes will be removed from the staging area. At commit time all staged
    dataframes will be concatenated. The committed data is one consolidated dataframe.

    Versioning:
    Committed data is versioned. At commit time a 'commit mode' can be provided to
    determine how the committed data is merged with the already committed data
    (e.g. overwrite, append, upsert, etc.).
    """

    @abc.abstractmethod
    def get_schema(
        self, which: Literal["data", "staging"] = "data"
    ) -> dict[str, pl.DataType]:
        """Get the schema for the specified data type.

        Parameters
        ----------
        which : Literal["data", "staging"], default="data"
            The type of schema to get.
            If "data" will get the schema for the committed data.
            If "staging" will get the schema for the staged data.

        Returns
        -------
        dict[str, pl.DataType]
            The schema of the uploaded files.
        """
        ...

    @abc.abstractmethod
    def get_summary_schema(
        self, which: Literal["data", "staging"] = "data"
    ) -> dict[str, pl.DataType]:
        """Get the summary schema for the specified data type.

        Parameters
        ----------
        which : Literal["data", "staging"], default="data"
            The type of summary statistics to get the schema for.
            If "data" will get the schema for the committed data.
            If "staging" will get the schema for the staged data.

        Returns
        -------
        dict[str, pl.DataType]
            The schema of the summary statistics.
        """
        ...

    @abc.abstractmethod
    def get_detail_summary_schema(
        self, which: Literal["data", "staging"] = "data"
    ) -> dict[str, pl.DataType]:
        """Get the detail summary schema for the specified data type.

        Parameters
        ----------
        which : Literal["data", "staging"], default="data"
            The type of detail summary statistics to get the schema for.
            If "data" will get the schema for the committed data.
            If "staging" will get the schema for the staged data.

        Returns
        -------
        dict[str, pl.DataType]
            The schema of the detail summary statistics.
        """
        ...

    @abc.abstractmethod
    def get_commit_modes(self) -> dict[str, str]:
        """Get the available commit modes and their descriptions.

        Returns
        -------
        dict[str, str]
            Available modes (keys) that can be passed to the commit function
            and their descriptions (values).
        """
        ...

    @abc.abstractmethod
    def get_parser_names(self) -> list[str]:
        """Get the list of available parser names.

        Returns
        -------
        list[str]
            The list of available parsers that can be used to parse files with
            this uploader.
        """
        ...

    @abc.abstractmethod
    def get_parser(self, parser: str) -> UploadParserApi:
        """Get a specific parser by name.

        Parameters
        ----------
        parser : str
            The parser to obtain.

        Returns
        -------
        UploadParserApi
            The parser.
        """
        ...

    @abc.abstractmethod
    def can_handle(
        self, df: pl.DataFrame, *, parser: str | None = None, name: str | None = None
    ) -> MultiParserResult:
        """Check if the uploader can handle the given dataframe.

        Parameters
        ----------
        df : pl.DataFrame
            The dataframe to check if the parser can handle.
        parser : str | None, default=None
            The parser to use. If None will check all parsers and choose the first
            parser that can handle the dataframe.
        name : str | None, default=None
            The name of the dataframe (which could be the filename).
            Some parsers might extract information (such as a date) from the name.

        Returns
        -------
        MultiParserResult
            The result of the parser check.
        """
        ...

    @abc.abstractmethod
    def stage_df(
        self,
        name: str,
        df: pl.DataFrame,
        parser: str | None = None,
        replace: bool = False,
    ) -> UploadStagingResult:
        """Stage a dataframe for upload under the given name.

        Parameters
        ----------
        name : str
            The name of the dataframe to stage.
        df : pl.DataFrame
            The dataframe to stage.
        parser : str | None, default=None
            The parser to use. If None will check all parsers.
        replace : bool, default=False
            If True will replace a possible existing staging with the same name.
            If False a name clash will result in a failed staging result.

        Raises
        ------
        UploadError
            If the given name already exists.

        Returns
        -------
        UploadStagingResult
            The result of the staging operation.
        """
        ...

    @abc.abstractmethod
    def stage_df_as_task(  # noqa: D102
        self,
        name: str,
        df: pl.DataFrame,
        parser: str | None = None,
        replace: bool = False,
    ) -> Task[UploadStagingResult]: ...

    @abc.abstractmethod
    def stage_file(
        self,
        path: Path,
        name: str | None = None,
        parser: str | None = None,
        replace: bool = False,
    ) -> UploadStagingResult:
        """Stage a file for upload under the given name.

        Parameters
        ----------
        path : Path
            File path for the file to stage.
        name : str | None, default=None
            The name of the dataframe to stage. If None will use the file name.
        parser : str | None, default=None
            The parser to use. If None will check all parsers.
        replace : bool, default=False
            If True will replace a possible existing staging with the same name.
            If False a name clash will result in a failed staging result.

        Raises
        ------
        UploadError
            If the file could not be staged.

        Returns
        -------
        UploadStagingResult
            The result of the staging operation.
        """
        ...

    @abc.abstractmethod
    def stage_file_as_task(  # noqa: D102
        self,
        path: Path,
        name: str | None = None,
        parser: str | None = None,
        replace: bool = False,
    ) -> Task[UploadStagingResult]: ...

    @abc.abstractmethod
    def get_staging_results(
        self, names: list[str] | None = None
    ) -> dict[str, UploadStagingResult]:
        """Get the staging results for the optional list of names.

        Parameters
        ----------
        names : list[str] | None, default=None
            The names of the staging results to get.
            If None will get all staging results.
            Missing names will be ignored and won't be part of the result.

        Returns
        -------
        dict[str, UploadStagingResult]
            A dictionary of staging results.
        """
        ...

    @abc.abstractmethod
    def wipe_staging(
        self,
        names: list[str] | None = None,
    ) -> dict[str, UploadStagingResult]:
        """Wipe the staging results for the optional list of names.

        Parameters
        ----------
        names : list[str] | None, default=None
            The names of the staging results to wipe.
            If None will wipe all staging results.
            Names that do not exist will be ignored.

        Returns
        -------
        dict[str, UploadStagingResult]
            A dictionary of staging results that were wiped.
        """
        ...

    @abc.abstractmethod
    def get_staging_data(
        self,
        names: list[str] | None = None,
        *,
        columns: list[str] | None = None,
        unique: bool = False,
    ) -> pl.LazyFrame:
        """Get the staging data for the optional list of names.

        Parameters
        ----------
        names : list[str] | None, default=None
            The names of the staging data to get.
            If None will get all staging data.
            Names that do not exist will be ignored.
        columns : list[str] | None, default=None
            The columns to get. If None will get all columns.
        unique : bool, default=False
            If True will return a unique set of rows.
            If False will return all rows.

        Returns
        -------
        pl.LazyFrame
            The staging data with an additional column `_name` that contains
            the staging name.
        """
        ...

    @abc.abstractmethod
    def get_staging_data_as_task(  # noqa: D102
        self,
        names: list[str] | None = None,
        *,
        columns: list[str] | None = None,
        unique: bool = False,
    ) -> Task[pl.LazyFrame]: ...

    @abc.abstractmethod
    def get_staging_data_summary(self, names: list[str] | None = None) -> pl.DataFrame:
        """Get the staging summary statistics for the optional list of names.

        Parameters
        ----------
        names : list[str] | None, default=None
            The names for which to get the staging summary statistics.
            If None will get all staging data.
            Names that do not exist will be ignored.

        Returns
        -------
        pl.DataFrame
            The staging summary statistics data which will contain one row
            per staged file.
            The schema matches `get_summary_schema` with an additional column
            `_name` (at the beginning) of type str which contains the name of the
            staged file.
        """
        ...

    @abc.abstractmethod
    def get_staging_data_summary_as_task(  # noqa: D102
        self, names: list[str] | None = None
    ) -> Task[pl.DataFrame]: ...

    @abc.abstractmethod
    def get_staging_data_detail_summary(
        self, names: list[str] | None = None
    ) -> pl.DataFrame:
        """Get the detail staging summary statistics for the optional list of names.

        Parameters
        ----------
        names : list[str] | None, default=None
            The names for which to get the detail staging summary statistics.
            If None will get all staging data.
            Names that do not exist will be ignored.

        Returns
        -------
        pl.DataFrame
            The detail staging summary statistics data which will contain multiple rows
            per staged file.
            The schema matches `get_detail_summary_schema` with an additional column
            `_name` (at the beginning) of type str which contains the name of the
            staged file.
        """
        ...

    @abc.abstractmethod
    def get_staging_data_detail_summary_as_task(  # noqa: D102
        self, names: list[str] | None = None
    ) -> Task[pl.DataFrame]: ...

    @abc.abstractmethod
    def validate_staging_data(
        self, names: list[str] | None = None, short: bool = False
    ) -> dict[str, pl.DataFrame]:
        """Validate the staging data for the optional list of names, and return errors.

        The errors are returned as a dict of error dataframes keyed by the name of the
        validation test.

        Parameters
        ----------
        names : list[str] | None, default=None
            The names for which to carry out the validation.
            If None will use all staging data.
            Names that do not exist will be ignored.
        short : bool, default=False
            If True will return a shorter dataframes which only contain a column `_name`
            to indicate the name of the staging file which has validation errors with
            other staging files.
            If False will return a dataframe with the `_name` (as above) and other
            columns to indicate the validation errors.

        Returns
        -------
        dict[str, pl.DataFrame]
            Error dataframes for different validation checks. The dataframes will
            contain a column `_name`. If `short` is False then the `_name` column
            contains a comma separated list of staging names that are the cause of the
            respective validation error.
            If `short` is True then the `_name` column will be the only column and
            each row contains a staging file name that has validation errors.
        """
        ...

    @abc.abstractmethod
    def validate_staging_data_as_task(  # noqa: D102
        self, names: list[str] | None = None, short: bool = False
    ) -> Task[dict[str, pl.DataFrame]]: ...

    @abc.abstractmethod
    def commit(self, mode: str, names: list[str] | None = None) -> UploadCommitResult:
        """Commit and clear all staged files identified by given names.

        Parameters
        ----------
        mode : str
            Commit mode, one of `commit_modes`.
        names : list[str] | None, default=None
            The names of the staging results to commit.
            If None will commit all staging results.
            Names that do not exist will be ignored.

        Raises
        ------
        ValueError
            If the given commit mode does not exist.
        UploadError
            If the staged data fails validation (as tested by `validate_staging_data`).

        Returns
        -------
        UploadCommitResult
            The result of the commit operation.
        """
        ...

    @abc.abstractmethod
    def commit_as_task(  # noqa: D102
        self, mode: str, names: list[str] | None = None
    ) -> Task[UploadCommitResult]: ...

    @abc.abstractmethod
    def fast_commit(
        self, df: pl.DataFrame | Path, mode: str, parser: str | None = None
    ) -> UploadCommitResult:
        """Directly commit the given dataframe without staging.

        Parameters
        ----------
        df : pl.DataFrame | Path
            The dataframe to commit.
        mode : str
            Commit mode, one of `commit_modes`.
        parser : str | None, default=None
            The parser to use. If None will check all parsers.

        Raises
        ------
        ValueError
            If the given commit mode does not exist.
        UploadError
            If the staged data fails validation (as tested by `validate_staging_data`).

        Returns
        -------
        UploadCommitResult
            The result of the commit operation.
        """
        ...

    @abc.abstractmethod
    def fast_commit_as_task(  # noqa: D102
        self, df: pl.DataFrame | Path, mode: str, parser: str | None = None
    ) -> Task[UploadCommitResult]: ...

    @abc.abstractmethod
    def get_data(
        self,
        columns: list[str] | None = None,
        *,
        filters: DNFFilterExpressions | None = None,
        unique: bool = False,
        head: int | None = None,
        version: int | dt.datetime | None = None,
        download_to: Path | str | None = None,
        download_filename: str = "data-{i}.parquet",
    ) -> pl.LazyFrame:
        """Get the committed data at the given version.

        Parameters
        ----------
        columns : list[str] | None, default=None
            The columns to get. If None will get all columns.
        filters : DNFFilterExpressions | None, default=None
            The filters to apply to the data. If None will get all data.
            Follow the pyarrow filter syntax.
            https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html
        unique : bool, default=False
            If True will return a unique set of rows.
            If False will return all rows.
        head : int | None, default=None
            If given will return the first `head` rows.
            If None will return all rows.
        version : int | dt.datetime | None, default=None
            The version of the underlying data, latest if not given.
            By definition version 0 is the empty dataframe with correct schema.
        download_to : Path | str | None, default=None
            If given will download the data to the given path and then return
            a LazyFrame against that downloaded path. The path must point to a
            an existing directory. If a string is given then a Path object will be
            created from it using `Path(download_to)`.
            If None will consume the data in memory and return a lazy frame.
        download_filename : str, default="data-{i}.parquet"
            The filename to use for the downloaded data which must contain the
            placeholder `{i}` which will be replaced with the index of the file.
            Only used if `download_to` is given.
            If `download_to/download_filename` already exists then an exception
            will be raised.

        Raises
        ------
        KeyError
            If the given version does not exist.
        FileExistsError
            If `download_to/download_filename` already exists.
        FileNotFoundError
            If `download_to` is given and does not exist.

        Returns
        -------
        pl.LazyFrame
            The committed data at the given version.
        """
        ...

    @abc.abstractmethod
    def get_data_as_task(  # noqa: D102
        self,
        columns: list[str] | None = None,
        *,
        filters: DNFFilterExpressions | None = None,
        unique: bool = False,
        head: int | None = None,
        version: int | dt.datetime | None = None,
        download_to: Path | str | None = None,
        download_filename: str = "data-{i}.parquet",
    ) -> Task[pl.LazyFrame]: ...

    @abc.abstractmethod
    def get_data_summary(
        self, version: int | dt.datetime | None = None
    ) -> pl.DataFrame:
        """Get the data summary statistics (single row) at the given version.

        Parameters
        ----------
        version : int | dt.datetime | None, default=None
            The version of the underlying data, latest if not given.

        Returns
        -------
        pl.DataFrame
            The summary statistics data (single row) for the data at the given
            version.
        """
        ...

    @abc.abstractmethod
    def get_data_summary_as_task(  # noqa: D102
        self, version: int | dt.datetime | None = None
    ) -> Task[pl.DataFrame]: ...

    @abc.abstractmethod
    def get_data_detail_summary(
        self, version: int | dt.datetime | None = None
    ) -> pl.DataFrame:
        """Get the detail data summary statistics (multi row) at the given version.

        Parameters
        ----------
        version : int | dt.datetime | None, default=None
            The version of the underlying data, latest if not given.

        Returns
        -------
        pl.DataFrame
            The summary statistics data (multi row) for the data at the given
            version.
        """
        ...

    @abc.abstractmethod
    def get_data_detail_summary_as_task(  # noqa: D102
        self, version: int | dt.datetime | None = None
    ) -> Task[pl.DataFrame]: ...

    @abc.abstractmethod
    def destroy(self) -> None:
        """Irreversibly delete the entire dataset.

        This includes the entire version history of uploads and staged files. Cannot be
        undone.
        """
        ...

    @abc.abstractmethod
    def version_history(self) -> dict[int, dt.datetime]:
        """Get the version history of the dataset.

        Returns
        -------
        dict[int, dt.datetime]
            A dictionary of version numbers and the corresponding timestamps.
        """
        ...


@docstrings_from_sync
class AsyncUploaderApi(abc.ABC):

    @abc.abstractmethod
    async def get_schema(  # noqa: D102
        self, which: Literal["data", "staging"] = "data"
    ) -> dict[str, pl.DataType]: ...

    @abc.abstractmethod
    async def get_summary_schema(  # noqa: D102
        self, which: Literal["data", "staging"] = "data"
    ) -> dict[str, pl.DataType]: ...

    @abc.abstractmethod
    async def get_detail_summary_schema(  # noqa: D102
        self, which: Literal["data", "staging"] = "data"
    ) -> dict[str, pl.DataType]: ...

    @abc.abstractmethod
    async def get_commit_modes(self) -> dict[str, str]: ...  # noqa: D102

    @abc.abstractmethod
    async def get_parser_names(self) -> list[str]: ...  # noqa: D102

    @abc.abstractmethod
    async def get_parser(self, parser: str) -> AsyncUploadParserApi: ...  # noqa: D102

    @abc.abstractmethod
    async def can_handle(  # noqa: D102
        self, df: pl.DataFrame, *, parser: str | None = None, name: str | None = None
    ) -> MultiParserResult: ...

    @abc.abstractmethod
    async def stage_df(  # noqa: D102
        self,
        name: str,
        df: pl.DataFrame,
        parser: str | None = None,
        replace: bool = False,
    ) -> UploadStagingResult: ...

    @abc.abstractmethod
    async def stage_df_as_task(  # noqa: D102
        self,
        name: str,
        df: pl.DataFrame,
        parser: str | None = None,
        replace: bool = False,
    ) -> AsyncTask[UploadStagingResult]: ...

    @abc.abstractmethod
    async def stage_file(  # noqa: D102
        self,
        path: Path,
        name: str | None = None,
        parser: str | None = None,
        replace: bool = False,
    ) -> UploadStagingResult: ...

    @abc.abstractmethod
    async def stage_file_as_task(  # noqa: D102
        self,
        path: Path,
        name: str | None = None,
        parser: str | None = None,
        replace: bool = False,
    ) -> AsyncTask[UploadStagingResult]: ...

    @abc.abstractmethod
    async def get_staging_results(  # noqa: D102
        self, names: list[str] | None = None
    ) -> dict[str, UploadStagingResult]: ...

    @abc.abstractmethod
    async def wipe_staging(  # noqa: D102
        self,
        names: list[str] | None = None,
    ) -> dict[str, UploadStagingResult]: ...

    @abc.abstractmethod
    async def get_staging_data(  # noqa: D102
        self,
        names: list[str] | None = None,
        *,
        columns: list[str] | None = None,
        unique: bool = False,
    ) -> pl.LazyFrame: ...

    @abc.abstractmethod
    async def get_staging_data_as_task(  # noqa: D102
        self,
        names: list[str] | None = None,
        *,
        columns: list[str] | None = None,
        unique: bool = False,
    ) -> AsyncTask[pl.LazyFrame]: ...

    @abc.abstractmethod
    async def get_staging_data_summary(  # noqa: D102
        self, names: list[str] | None = None
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def get_staging_data_summary_as_task(  # noqa: D102
        self, names: list[str] | None = None
    ) -> AsyncTask[pl.DataFrame]: ...

    @abc.abstractmethod
    async def get_staging_data_detail_summary(  # noqa: D102
        self, names: list[str] | None = None
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def get_staging_data_detail_summary_as_task(  # noqa: D102
        self, names: list[str] | None = None
    ) -> AsyncTask[pl.DataFrame]: ...

    @abc.abstractmethod
    async def validate_staging_data(  # noqa: D102
        self, names: list[str] | None = None, short: bool = False
    ) -> dict[str, pl.DataFrame]: ...

    @abc.abstractmethod
    async def validate_staging_data_as_task(  # noqa: D102
        self, names: list[str] | None = None, short: bool = False
    ) -> AsyncTask[dict[str, pl.DataFrame]]: ...

    @abc.abstractmethod
    async def commit(  # noqa: D102
        self, mode: str, names: list[str] | None = None
    ) -> UploadCommitResult: ...

    @abc.abstractmethod
    async def commit_as_task(  # noqa: D102
        self, mode: str, names: list[str] | None = None
    ) -> AsyncTask[UploadCommitResult]: ...

    @abc.abstractmethod
    async def fast_commit(  # noqa: D102
        self, df: pl.DataFrame | Path, mode: str, parser: str | None = None
    ) -> UploadCommitResult: ...

    @abc.abstractmethod
    async def fast_commit_as_task(  # noqa: D102
        self, df: pl.DataFrame | Path, mode: str, parser: str | None = None
    ) -> AsyncTask[UploadCommitResult]: ...

    @abc.abstractmethod
    async def get_data(  # noqa: D102
        self,
        columns: list[str] | None = None,
        *,
        filters: DNFFilterExpressions | None = None,
        unique: bool = False,
        head: int | None = None,
        version: int | dt.datetime | None = None,
        download_to: Path | str | None = None,
        download_filename: str = "data-{i}.parquet",
    ) -> pl.LazyFrame: ...

    @abc.abstractmethod
    async def get_data_as_task(  # noqa: D102
        self,
        columns: list[str] | None = None,
        *,
        filters: DNFFilterExpressions | None = None,
        unique: bool = False,
        head: int | None = None,
        version: int | dt.datetime | None = None,
        download_to: Path | str | None = None,
        download_filename: str = "data-{i}.parquet",
    ) -> AsyncTask[pl.LazyFrame]: ...

    @abc.abstractmethod
    async def get_data_summary(  # noqa: D102
        self, version: int | dt.datetime | None = None
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def get_data_summary_as_task(  # noqa: D102
        self, version: int | dt.datetime | None = None
    ) -> AsyncTask[pl.DataFrame]: ...

    @abc.abstractmethod
    async def get_data_detail_summary(  # noqa: D102
        self, version: int | dt.datetime | None = None
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def get_data_detail_summary_as_task(  # noqa: D102
        self, version: int | dt.datetime | None = None
    ) -> AsyncTask[pl.DataFrame]: ...

    @abc.abstractmethod
    async def destroy(self) -> None: ...  # noqa: D102

    @abc.abstractmethod
    async def version_history(self) -> dict[int, dt.datetime]: ...  # noqa: D102


class DataTypeUploaderApi(abc.ABC):
    """Expose different datasets of an uploader.

    Datasets are completely isolated spaces to upload data of the same type.
    """

    @abc.abstractmethod
    def get_datasets(self) -> list[str]:
        """Get the list of available datasets.

        Returns
        -------
        list[str]
            A sorted list of available datasets.
        """
        ...

    @abc.abstractmethod
    def get_dataset(self, dataset: str) -> UploaderApi:
        """Get a specific dataset uploader.

        Parameters
        ----------
        dataset : str
            The dataset of the uploader to obtain.
            Must be one of the datasets returned by `get_datasets`.

        Raises
        ------
        KeyError
            If the given dataset does not exist.

        Returns
        -------
        UploaderApi
            The uploader for the specified dataset.
        """
        ...

    @abc.abstractmethod
    def create_dataset(self, dataset: str) -> UploaderApi:
        """Create a new dataset.

        Parameters
        ----------
        dataset : str
            The dataset to create.

        Raises
        ------
        ValueError
            If the given dataset already exists.

        Returns
        -------
        UploaderApi
            The uploader for the new dataset.
        """
        ...

    def get_or_create_dataset(self, dataset: str) -> UploaderApi:
        """Get the dataset if it exists, otherwise create it.

        Parameters
        ----------
        dataset : str
            The dataset to get or create.

        Returns
        -------
        UploaderApi
            The uploader for the dataset.
        """
        if dataset in self.get_datasets():
            return self.get_dataset(dataset)
        else:
            return self.create_dataset(dataset)

    def create_or_replace_dataset(self, dataset: str) -> UploaderApi:
        """Create a new dataset and delete the old one if it exists.

        Parameters
        ----------
        dataset : str
            The dataset to create.

        Returns
        -------
        UploaderApi
            The uploader for the dataset.
        """
        if dataset in self.get_datasets():
            self.get_dataset(dataset).destroy()

        return self.create_dataset(dataset)


@docstrings_from_sync
class AsyncDataTypeUploaderApi(abc.ABC):

    @abc.abstractmethod
    async def get_datasets(self) -> list[str]: ...  # noqa: D102

    @abc.abstractmethod
    async def get_dataset(self, dataset: str) -> AsyncUploaderApi: ...  # noqa: D102

    @abc.abstractmethod
    async def create_dataset(self, dataset: str) -> AsyncUploaderApi: ...  # noqa: D102

    async def get_or_create_dataset(  # noqa: D102
        self, dataset: str
    ) -> AsyncUploaderApi:
        if dataset in await self.get_datasets():
            return await self.get_dataset(dataset)
        else:
            return await self.create_dataset(dataset)

    async def create_or_replace_dataset(  # noqa: D102
        self, dataset: str
    ) -> AsyncUploaderApi:
        if dataset in await self.get_datasets():
            await (await self.get_dataset(dataset)).destroy()

        return await self.create_dataset(dataset)


class UploadersApi(abc.ABC):
    """Expose uploaders for different data types, e.g. exposures, portfolios, etc."""

    @abc.abstractmethod
    def get_data_types(self) -> list[str]:
        """Get the list of available data types.

        Returns
        -------
        list[str]
            A sorted list of available upload data types,
            e.g. "exposures", "portfolios", etc.
        """
        ...

    @abc.abstractmethod
    def get_data_type(self, data_type: str) -> DataTypeUploaderApi:
        """Get a specific data type uploader.

        Parameters
        ----------
        data_type : str
            The data type of the uploader to obtain.
            One of the data types returned by `get_data_types`.

        Raises
        ------
        KeyError
            If the given data type does not exist.

        Returns
        -------
        DataTypeUploaderApi
            The uploader for the given data type.
        """
        ...


@docstrings_from_sync
class AsyncUploadersApi(abc.ABC):

    @abc.abstractmethod
    async def get_data_types(self) -> list[str]: ...  # noqa: D102

    @abc.abstractmethod
    async def get_data_type(  # noqa: D102
        self, data_type: str
    ) -> AsyncDataTypeUploaderApi: ...
